
from __future__ import annotations

import abc
import asyncio
import copy
import functools
import inspect
import logging

from typing import Any, Dict, List, Optional, Set, Union

from qtoggleserver import persist
from qtoggleserver.conf import settings
from qtoggleserver.core import events as core_events
from qtoggleserver.core import expressions as core_expressions
from qtoggleserver.core import history as core_history
from qtoggleserver.core import main
from qtoggleserver.core import sequences as core_sequences
from qtoggleserver.core.typing import Attribute, Attributes, AttributeDefinitions, GenericJSONDict
from qtoggleserver.core.typing import NullablePortValue, PortValue
from qtoggleserver.utils import dynload as dynload_utils
from qtoggleserver.utils import json as json_utils
from qtoggleserver.utils import logging as logging_utils


TYPE_BOOLEAN = 'boolean'
TYPE_NUMBER = 'number'

CHANGE_REASON_NATIVE = 'N'
CHANGE_REASON_API = 'A'
CHANGE_REASON_SEQUENCE = 'S'
CHANGE_REASON_EXPRESSION = 'E'

logger = logging.getLogger(__name__)

_ports_by_id: Dict[str, BasePort] = {}


async def _attrdef_unit_enabled(port: BasePort) -> bool:
    return await port.get_type() == TYPE_NUMBER


STANDARD_ATTRDEFS = {
    'id': {
        'type': 'string'
    },
    'display_name': {
        'type': 'string',
        'modifiable': True,
        'max': 64
    },
    'type': {
        'type': 'string',
        'choices': [
            {
                'value': 'boolean',
                'display_name': 'Boolean'
            },
            {
                'value': 'number',
                'display_name': 'Number'
            }
        ]
    },
    'unit': {
        'type': 'string',
        'modifiable': True,
        'max': 16,
        'enabled': _attrdef_unit_enabled
    },
    'writable': {
        'type': 'boolean'
    },
    'enabled': {
        'type': 'boolean',
        'modifiable': True
    },

    'min': {
        'type': 'number',
        'optional': True
    },
    'max': {
        'type': 'number',
        'optional': True
    },
    'integer': {
        'type': 'boolean',
        'optional': True
    },
    'step': {
        'type': 'number',
        'optional': True
    },
    'choices': {  # TODO data type uncertain
        'type': '[]',
        'optional': True
    },

    'tag': {
        'type': 'string',
        'optional': True,
        'modifiable': True,
        'max': 64,
    },
    'expression': {
        'type': 'string',
        'optional': True,
        'modifiable': True,
        'max': 1024,
    },
    'transform_read': {
        'type': 'string',
        'optional': True,
        'modifiable': True,
        'max': 1024,
    },
    'transform_write': {
        'type': 'string',
        'optional': True,
        'modifiable': True,
        'max': 1024,
    },
    'persisted': {
        'type': 'boolean',
        'optional': True,
        'modifiable': True
    },
    'internal': {
        'type': 'boolean',
        'optional': True,
        'modifiable': True
    },
    'virtual': {
        'type': 'boolean',
        'optional': True
    },
    'online': {
        'type': 'boolean',
        'optional': True
    },
    'history_interval': {
        'type': 'number',
        'integer': True,
        'min': -1,
        'max': 2147483647,
        'optional': True,
        'modifiable': True,
        'enabled': lambda p: core_history.is_enabled()
    },
    'history_retention': {
        'type': 'number',
        'integer': True,
        'min': 0,
        'max': 2147483647,
        'optional': True,
        'modifiable': True,
        'enabled': lambda p: core_history.is_enabled()
    }
}


class PortError(Exception):
    pass


class PortLoadError(PortError):
    pass


class PortReadError(PortError):
    pass


class InvalidAttributeValue(PortError):
    def __init__(self, attr: str, details: Optional[GenericJSONDict] = None) -> None:
        self.attr: str = attr
        self.details: Optional[GenericJSONDict] = details

        super().__init__(attr)


class PortTimeout(PortError):
    pass


class BasePort(logging_utils.LoggableMixin, metaclass=abc.ABCMeta):
    PERSIST_COLLECTION = 'ports'

    TYPE = TYPE_BOOLEAN
    DISPLAY_NAME = ''
    UNIT = ''
    WRITABLE = False
    CHOICES = None

    WRITE_VALUE_QUEUE_SIZE = 1024

    STANDARD_ATTRDEFS = STANDARD_ATTRDEFS

    ADDITIONAL_ATTRDEFS = {}
    '''
    ADDITIONAL_ATTRDEFS = {
        'attr1': {
            'display_name': 'Some Attribute Display Name',
            'description': 'Some attribute description',
            'type': 'number',
            'modifiable': True,
            'pattern': '^.*$',
            'unit': 'seconds',
            'min': 1,
            'max': 100,
            'integer': True,
            'step': 5,
            'choices': [
                {'value': 2, 'display_name': 'Two'},
                {'value': 4, 'display_name': 'Four'}
            ]
        },
        ...
    }'''

    def __init__(self, port_id: str) -> None:
        logging_utils.LoggableMixin.__init__(self, port_id, logger)

        self._id: str = port_id
        self._enabled: bool = False
        self._display_name: Optional[str] = None
        self._unit: Optional[str] = None

        self._sequence: Optional[core_sequences.Sequence] = None
        self._expression: Optional[core_expressions.Expression] = None
        self._transform_read: Optional[core_expressions.Expression] = None
        self._transform_write: Optional[core_expressions.Expression] = None

        self._history_interval: int = 0
        self._history_retention: int = 0
        self._history_last_timestamp: int = 0

        # Attributes cache is used to prevent computing an attribute value more than once per core iteration
        self._attrs_cache: Attributes = {}

        # Cache attribute definitions so that the ATTRDEFS property doesn't need to gather all of them on each access
        self._attrdefs_cache: Optional[AttributeDefinitions] = None

        self._modifiable_attrs: Optional[Set[str]] = None
        self._non_modifiable_attrs: Optional[Set[str]] = None

        self._schema: Optional[GenericJSONDict] = None
        self._value_schema: Optional[GenericJSONDict] = None

        self._last_read_value: NullablePortValue = None
        self._write_value_queue: asyncio.Queue = asyncio.Queue(maxsize=self.WRITE_VALUE_QUEUE_SIZE)
        self._write_value_task: Optional[asyncio.Task] = None
        if asyncio.get_event_loop().is_running():
            self._write_value_task = asyncio.create_task(self._write_value_loop())
        self._change_reason: str = CHANGE_REASON_NATIVE
        self._pending_read: bool = False
        self._pending_write: bool = False

        self._save_lock: asyncio.Lock = asyncio.Lock()

        self._loaded: bool = False

    def __str__(self) -> str:
        return f'port {self._id}'

    def __repr__(self) -> str:
        return f'Port({self._id})'

    def initialize(self) -> None:
        pass

    async def get_attrdefs(self) -> AttributeDefinitions:
        if self._attrdefs_cache is None:
            self._attrdefs_cache = dict(self.STANDARD_ATTRDEFS, **self.ADDITIONAL_ATTRDEFS)

            for name, attrdef in list(self._attrdefs_cache.items()):
                enabled = attrdef.get('enabled', True)
                if callable(enabled):
                    enabled = enabled(self)
                if inspect.isawaitable(enabled):
                    enabled = await enabled

                if not enabled:
                    self._attrdefs_cache.pop(name)

        return self._attrdefs_cache

    def invalidate_attrdefs(self) -> None:
        self._attrs_cache = {}
        self._attrdefs_cache = None
        self._modifiable_attrs = None
        self._non_modifiable_attrs = None
        self._schema = None

    async def get_non_modifiable_attrs(self) -> Set[str]:
        attrdefs = await self.get_attrdefs()

        if self._non_modifiable_attrs is None:
            self._non_modifiable_attrs = set(n for (n, v) in attrdefs.items() if not v.get('modifiable'))

        return self._non_modifiable_attrs

    async def get_modifiable_attrs(self) -> Set[str]:
        attrdefs = await self.get_attrdefs()

        if self._modifiable_attrs is None:
            self._modifiable_attrs = set(n for (n, v) in attrdefs.items() if v.get('modifiable'))

        return self._modifiable_attrs

    async def get_attrs(self) -> Attributes:
        d = {}

        for name in await self.get_attrdefs():
            v = await self.get_attr(name)
            if v is None:
                continue

            d[name] = v

        return d

    def invalidate_attrs(self) -> None:
        self._attrs_cache = {}

    async def get_attr(self, name: str) -> Optional[Attribute]:
        value = self._attrs_cache.get(name)
        if value is not None:
            return value

        method = getattr(self, 'attr_get_' + name, getattr(self, 'attr_is_' + name, None))
        if method:
            value = await method()
            if value is not None:
                self._attrs_cache[name] = value
                return value

        value = getattr(self, '_' + name, None)
        if value is not None:
            self._attrs_cache[name] = value
            return value

        method = getattr(self, 'attr_get_default_' + name, getattr(self, 'attr_is_default_' + name, None))
        if method:
            value = await method()
            if value is not None:
                self._attrs_cache[name] = value
                return value

        value = getattr(self, name.upper(), None)
        if value is not None:
            self._attrs_cache[name] = value
            return value

        return None  # Unsupported attribute

    async def set_attr(self, name: str, value: Attribute) -> None:
        old_value = await self.get_attr(name)
        if old_value is None:
            return  # Refuse to set an unsupported attribute

        method = getattr(self, 'attr_set_' + name, None)
        if method:
            try:
                await method(value)

            except Exception:
                self.error('failed to set attribute %s = %s', name, json_utils.dumps(value), exc_info=True)

                raise

        elif hasattr(self, '_' + name):
            setattr(self, '_' + name, value)

        if self.is_loaded():
            await main.update()

        # New attributes might have been added or removed after setting an attribute; therefore new definitions might
        # have appeared or disappeared
        self.invalidate_attrdefs()

        self._attrs_cache[name] = value
        if old_value != value:
            await self.handle_attr_change(name, value)

        if not self.is_loaded():
            return

        # Skip an IO loop iteration, allowing setting multiple attributes before triggering a port-update
        await asyncio.sleep(0)
        await self.trigger_update()

    def invalidate_attr(self, name: str) -> None:
        self._attrs_cache.pop(name, None)

    async def handle_attr_change(self, name: str, value: Attribute) -> None:
        method_name = f'handle_{name}'
        method = getattr(self, method_name, None)
        if method:
            try:
                await method(value)

            except Exception as e:
                self.error('%s failed: %s', method_name, e, exc_info=True)

    async def get_display_name(self) -> str:
        return await self.get_attr('display_name') or self._id

    async def get_display_value(self, value: NullablePortValue = None) -> str:
        choices = await self.get_attr('choices')
        unit = await self.get_attr('unit')
        if value is None:
            value = self.get_last_read_value()

        if value is None:
            return 'unknown'  # TODO: i18n

        if choices:
            for choice in choices:
                if choice['value'] == value:
                    return choice['display_name']

        if await self.get_type() == TYPE_BOOLEAN:
            value_str = 'on' if value else 'off'  # TODO: i18n

        else:
            value_str = str(value)

        if unit:
            return f'{value_str}{unit}'

        else:
            return value_str

    def get_id(self) -> str:
        return self._id

    def map_id(self, new_id: str) -> None:
        self._id = new_id
        self.debug('mapped to %s', new_id)
        self.set_logger_name(new_id)

    async def get_type(self) -> str:
        return await self.get_attr('type')

    async def is_writable(self) -> bool:
        return await self.get_attr('writable')

    async def is_persisted(self) -> bool:
        return await self.get_attr('persisted')

    async def is_internal(self) -> bool:
        return await self.get_attr('internal')

    def is_enabled(self) -> bool:
        return self._enabled

    async def enable(self) -> None:
        if self._enabled:
            return

        self.debug('enabling')
        self._enabled = True
        self.invalidate_attr('enabled')

        # Reset port expression
        if self._expression:
            sexpression = str(self._expression)
            self.debug('resetting expression "%s"', sexpression)
            self._expression = core_expressions.parse(self.get_id(), sexpression)

            main.force_eval_expressions(self)

        try:
            await self.handle_enable()

        except Exception:
            self.error('failed to enable')
            self._enabled = False

            raise

    async def disable(self) -> None:
        if not self._enabled:
            return

        # Cancel sequence
        if self._sequence:
            self.debug('canceling current sequence')
            await self._sequence.cancel()
            self._sequence = None

        self.debug('disabling')
        self._enabled = False
        self.invalidate_attr('enabled')

        try:
            await self.handle_disable()

        except Exception:
            self.error('failed to disable')
            self._enabled = True

            raise

    async def handle_enable(self) -> None:
        pass

    async def handle_disable(self) -> None:
        pass

    async def attr_set_enabled(self, value: bool) -> None:
        if value:
            await self.enable()

        else:
            await self.disable()

    async def get_expression(self) -> Optional[core_expressions.Expression]:
        if not await self.is_writable():
            return None

        return self._expression

    async def attr_get_expression(self) -> Optional[str]:
        if not await self.is_writable():
            return None

        if self._expression:
            return str(self._expression)

        else:
            return ''

    async def attr_set_expression(self, sexpression: str) -> None:
        if not await self.is_writable():
            self.error('refusing to set expression on non-writable port')
            raise PortError('Cannot set expression on non-writable port')

        if self._sequence:
            self.debug('canceling current sequence')
            await self._sequence.cancel()
            self._sequence = None

        if not sexpression:
            self._expression = None
            return

        try:
            self.debug('parsing expression "%s"', sexpression)
            expression = core_expressions.parse(self.get_id(), sexpression)

            self.debug('checking for expression circular dependencies')
            await core_expressions.check_loops(self, expression)

        except core_expressions.ExpressionParseError as e:
            self.error('failed to set expression "%s": %s', sexpression, e)

            raise InvalidAttributeValue('expression', details=e.to_json()) from e

        self.debug('setting expression "%s"', expression)
        self._expression = expression

        main.force_eval_expressions(self)

    async def attr_get_transform_read(self) -> str:
        if self._transform_read:
            return str(self._transform_read)

        else:
            return ''

    async def attr_set_transform_read(self, stransform_read: str) -> None:
        if not stransform_read:
            self._transform_read = None
            return

        try:
            self.debug('parsing expression "%s"', stransform_read)
            transform_read = core_expressions.parse(self.get_id(), stransform_read)

            deps = transform_read.get_deps()
            for dep in deps:
                if not dep.startswith('$'):
                    continue

                if dep != f'${self._id}':
                    raise core_expressions.ExternalDependency(
                        port_id=dep[1:],
                        pos=stransform_read.index(dep)
                    )

            self.debug('setting read transform "%s"', transform_read)
            self._transform_read = transform_read

        except core_expressions.ExpressionParseError as e:
            self.error('failed to set transform read expression "%s": %s', stransform_read, e)

            raise InvalidAttributeValue('transform_read', details=e.to_json()) from e

    async def attr_get_transform_write(self) -> Optional[str]:
        if not await self.is_writable():
            return None  # Only writable ports have transform_write attributes

        if self._transform_write:
            return str(self._transform_write)

        else:
            return ''

    async def attr_set_transform_write(self, stransform_write: str) -> None:
        if not stransform_write:
            self._transform_write = None
            return

        try:
            self.debug('parsing expression "%s"', stransform_write)
            transform_write = core_expressions.parse(self.get_id(), stransform_write)

            deps = transform_write.get_deps()
            for dep in deps:
                if not dep.startswith('$'):
                    continue

                if dep != f'${self._id}':
                    raise core_expressions.ExternalDependency(
                        port_id=dep[1:],
                        pos=stransform_write.index(dep)
                    )

            self.debug('setting write transform "%s"', transform_write)
            self._transform_write = transform_write

        except core_expressions.ExpressionParseError as e:
            self.error('failed to set transform write expression "%s": %s', stransform_write, e)

            raise InvalidAttributeValue('transform_write', details=e.to_json()) from e

    async def get_history_interval(self) -> int:
        return await self.get_attr('history_interval')

    async def get_history_retention(self) -> int:
        return await self.get_attr('history_retention')

    def get_history_last_timestamp(self) -> int:
        return self._history_last_timestamp

    def set_history_last_timestamp(self, timestamp: int) -> None:
        self._history_last_timestamp = timestamp

    @abc.abstractmethod
    async def read_value(self) -> NullablePortValue:
        return None

    async def write_value(self, value: PortValue) -> None:
        pass

    def get_last_read_value(self) -> NullablePortValue:
        return self._last_read_value

    def set_last_read_value(self, value: NullablePortValue) -> None:
        self._last_read_value = value

    def get_change_reason(self) -> str:
        return self._change_reason

    def reset_change_reason(self) -> None:
        self._change_reason = CHANGE_REASON_NATIVE

    async def read_transformed_value(self) -> NullablePortValue:
        if self._pending_read:
            return  # Prevent overlapped readings

        self._pending_read = True

        try:
            value = await self.read_value()

        except Exception:
            raise

        finally:
            self._pending_read = False

        if value is None:
            return None

        if self._transform_read:
            # Temporarily set the new value to the port, so that the read transform expression works as expected
            old_value = self._last_read_value
            self._last_read_value = value
            value = await self.adapt_value_type(await self._transform_read.eval())
            self._last_read_value = old_value

        return value

    def is_pending_read(self) -> bool:
        return self._pending_read

    async def write_transformed_value(self, value: PortValue, reason: str) -> None:
        if self._transform_write:
            # Temporarily set the port value to the new value, so that the write transform expression takes the new
            # value into consideration when evaluating the result
            prev_value = self._last_read_value
            self._last_read_value = value
            value = await self.adapt_value_type(await self._transform_write.eval())
            self._last_read_value = prev_value

        try:
            await self._write_value_queued(value, reason)
            self.debug('wrote value %s (reason=%s)', json_utils.dumps(value), reason)

        except Exception:
            self.error('failed to write value %s (reason=%s)', json_utils.dumps(value), reason, exc_info=True)

            raise

    def is_pending_write(self) -> bool:
        return self._pending_write

    def push_value(self, value: PortValue, reason: str) -> None:
        asyncio.create_task(self.write_transformed_value(value, reason))

    async def _write_value_queued(self, value: PortValue, reason: str) -> None:
        done = asyncio.get_running_loop().create_future()

        while True:
            try:
                self._write_value_queue.put_nowait((value, reason, done))

            except asyncio.QueueFull as e:
                self.warning('write queue full, dropping oldest value')

                # Reject the future of the dropped request with a QueueFull exception
                dropped = self._write_value_queue.get_nowait()
                dropped[2].set_exception(e)

            else:
                break

        # Wait for actual write_value operation to be done
        await done

    async def _write_value_loop(self) -> None:
        while True:
            try:
                value, reason, done = await self._write_value_queue.get()

                self._change_reason = reason
                self._pending_write = True

                try:
                    result = await self.write_value(value)
                    done.set_result(result)

                except Exception as e:
                    done.set_exception(e)

                self._pending_write = False

                await main.update()  # Do an update after every confirmed write

            except asyncio.CancelledError:
                self.debug('write value task cancelled')
                break

    async def adapt_value_type(self, value: NullablePortValue) -> NullablePortValue:
        return self.adapt_value_type_sync(await self.get_type(), await self.get_attr('integer'), value)

    @staticmethod
    def adapt_value_type_sync(type_: str, integer: bool, value: NullablePortValue) -> NullablePortValue:
        if value is None:
            return None

        if type_ == TYPE_BOOLEAN:
            return bool(value)

        else:
            # Round the value if port accepts only integers
            if integer:
                return int(value)

            return float(value)

    async def set_sequence(self, values: List[PortValue], delays: List[int], repeat: int) -> None:
        if self._sequence:
            self.debug('canceling current sequence')
            await self._sequence.cancel()
            self._sequence = None

        if values:
            callback = functools.partial(self._write_transformed_value_fire_and_forget, reason=CHANGE_REASON_SEQUENCE)
            self._sequence = core_sequences.Sequence(values, delays, repeat, callback, self._on_sequence_finish)

            self.debug('installing sequence')
            self._sequence.start()

    def _write_transformed_value_fire_and_forget(self, value: PortValue, reason: str) -> None:
        asyncio.create_task(self.write_transformed_value(value, reason))

    async def _on_sequence_finish(self) -> None:
        self.debug('sequence finished')

        self._sequence = None
        if await self.is_persisted():
            await self.save()

    def heart_beat(self) -> None:
        pass

    def heart_beat_second(self) -> None:
        pass

    async def to_json(self) -> GenericJSONDict:
        attrs = await self.get_attrs()
        attrs = dict(attrs)

        if self._enabled:
            attrs['value'] = self._last_read_value

        else:
            attrs['value'] = None

        attrdefs = copy.deepcopy(self.ADDITIONAL_ATTRDEFS)
        for attrdef in attrdefs.values():
            attrdef.pop('pattern', None)

        attrs['definitions'] = attrdefs

        return attrs

    async def load(self) -> None:
        self.debug('loading persisted data')

        data = await persist.get(self.PERSIST_COLLECTION, self.get_id()) or {}
        await self.load_from_data(data)

        self.set_loaded()
        self.initialize()

    async def reset(self) -> None:
        self.debug('resetting persisted data')
        await self.load_from_data(data={})
        self.invalidate_attrdefs()

    async def load_from_data(self, data: GenericJSONDict) -> None:
        attrs_start = ['enabled']  # These will be loaded first, in this order
        attrs_end = ['expression']  # These will be loaded last, in this order

        attr_items = data.items()
        attr_items = [a for a in attr_items if (a[0] not in attrs_start) and (a[0] not in attrs_end)]

        attr_items_start = []
        for n in attrs_start:
            v = data.get(n)
            if v is not None:
                attr_items_start.append((n, v))

        # Sort the rest of the attributes alphabetically
        attr_items.sort(key=lambda i: i[0])

        attr_items_end = []
        for n in attrs_end:
            v = data.get(n)
            if v is not None:
                attr_items_end.append((n, v))

        attr_items = attr_items_start + attr_items + attr_items_end

        for name, value in attr_items:
            if name in ('id', 'value'):
                continue  # Value is also among the persisted fields

            try:
                self.debug('loading %s = %s', name, json_utils.dumps(value))
                await self.set_attr(name, value)

            except Exception as e:
                self.error('failed to set attribute %s = %s: %s', name, json_utils.dumps(value), e)

        # value
        if await self.is_persisted() and data.get('value') is not None:
            self._last_read_value = data['value']
            self.debug('loaded value = %s', json_utils.dumps(self._last_read_value))

            if await self.is_writable():
                # Write the just-loaded value to the port
                value = self._last_read_value
                if self._transform_write:
                    value = await self.adapt_value_type(await self._transform_write.eval())

                await self.write_value(value)

        elif self.is_enabled():
            try:
                value = await self.read_transformed_value()
                if value is not None:
                    self._last_read_value = value
                    self.debug('read value = %s', json_utils.dumps(self._last_read_value))

            except Exception as e:
                self.error('failed to read value: %s', e, exc_info=True)

        # various
        self._history_last_timestamp = data.get('history_last_timestamp', 0)

    async def save(self) -> None:
        if not self.is_loaded():
            return

        self.debug('persisting data')

        async with self._save_lock:
            d = await self.prepare_for_save()
            await persist.replace(self.PERSIST_COLLECTION, self._id, d)

    async def prepare_for_save(self) -> GenericJSONDict:
        # value
        d = {
            'id': self.get_id(),
            'history_last_timestamp': self._history_last_timestamp
        }

        if await self.is_persisted():
            d['value'] = self._last_read_value

        else:
            d['value'] = None

        # attributes
        for name in await self.get_modifiable_attrs():
            v = await self.get_attr(name)
            if v is None:
                continue

            d[name] = v

        return d

    async def cleanup(self) -> None:
        if self._write_value_task:
            self._write_value_task.cancel()
            await self._write_value_task
            self._write_value_task = None

    def is_loaded(self) -> bool:
        return self._loaded

    def set_loaded(self) -> None:
        self._loaded = True

    async def remove(self, persisted_data: bool = True) -> None:
        await self.cleanup()

        self.debug('removing port')
        _ports_by_id.pop(self._id, None)

        if persisted_data:
            self.debug('removing persisted data')
            await persist.remove(self.PERSIST_COLLECTION, filt={'id': self._id})
            if core_history.is_enabled():
                await core_history.remove_samples([self], background=True)

        await self.trigger_remove()

    async def trigger_add(self) -> None:
        await core_events.trigger(core_events.PortAdd(self))

    async def trigger_remove(self) -> None:
        await core_events.trigger(core_events.PortRemove(self))

    async def trigger_update(self) -> None:
        await core_events.trigger(core_events.PortUpdate(self))

    async def trigger_value_change(self) -> None:
        await core_events.trigger(core_events.ValueChange(self))

    async def get_schema(self) -> GenericJSONDict:
        if self._schema is None:
            self._schema = {
                'type': 'object',
                'properties': {},
                'additionalProperties': False
            }

            attrdefs = await self.get_attrdefs()
            for name, attrdef in attrdefs.items():
                if not attrdef.get('modifiable'):
                    continue

                if await self.get_attr(name) is None:
                    continue

                attr_schema = dict(attrdef)

                if attr_schema['type'] == 'string':
                    if 'min' in attr_schema:
                        attr_schema['minLength'] = attr_schema.pop('min')

                    if 'max' in attr_schema:
                        attr_schema['maxLength'] = attr_schema.pop('max')

                elif attr_schema['type'] == 'number':
                    if attr_schema.get('integer'):
                        attr_schema['type'] = 'integer'

                    if 'min' in attr_schema:
                        attr_schema['minimum'] = attr_schema.pop('min')

                    if 'max' in attr_schema:
                        attr_schema['maximum'] = attr_schema.pop('max')

                if 'choices' in attrdef:
                    attr_schema['enum'] = [c['value'] for c in attr_schema.pop('choices')]

                attr_schema.pop('modifiable', None)

                self._schema['properties'][name] = attr_schema

        return self._schema

    async def get_value_schema(self) -> GenericJSONDict:
        if self._value_schema is None:
            self._value_schema = {}

            c = await self.get_attr('choices')
            if c is not None:
                self._value_schema['enum'] = [i['value'] for i in c]

            else:
                m = await self.get_attr('min')
                if m is not None:
                    self._value_schema['minimum'] = m

                m = await self.get_attr('max')
                if m is not None:
                    self._value_schema['maximum'] = m

                if await self.get_attr('integer'):
                    self._value_schema['type'] = 'integer'

                elif await self.get_type() == TYPE_BOOLEAN:
                    self._value_schema['type'] = 'boolean'

                else:  # Assuming number
                    self._value_schema['type'] = 'number'

        return self._value_schema


class Port(BasePort, metaclass=abc.ABCMeta):
    def __init__(self, port_id: str) -> None:
        super().__init__(port_id)

        self._tag: str = ''
        self._persisted: bool = False
        self._internal: bool = False


async def load(port_args: List[Dict[str, Any]], raise_on_error: bool = True) -> List[BasePort]:
    port_driver_classes = {}
    ports = []

    # Create ports
    for ps in port_args:
        driver = ps.pop('driver', None)
        if not driver:
            if raise_on_error:
                raise PortLoadError('Missing port driver')

            logger.error('ignoring port with no driver')
            continue

        if isinstance(driver, str):
            if driver not in port_driver_classes:
                try:
                    logger.debug('loading port driver %s', driver)
                    port_driver_classes[driver] = dynload_utils.load_attr(driver)

                except Exception as e:
                    if raise_on_error:
                        raise PortLoadError(f'Failed to load port driver {driver}') from e

                    logger.error('failed to load port driver %s: %s', driver, e, exc_info=True)
                    continue

            port_class = port_driver_classes[driver]

        else:
            port_class = driver

        port_class_desc = f'{port_class.__module__}.{port_class.__name__}'

        try:
            port = port_class(**ps)
            port_id = port.get_id()

            if port_id in _ports_by_id:
                raise PortLoadError(f'A port with id {port_id} already exists')

            _ports_by_id[port.get_id()] = port
            ports.append(port)

            logger.debug('initialized %s (driver %s)', port, port_class_desc)

        except Exception as e:
            if raise_on_error:
                raise PortLoadError(f'Failed to initialize port from driver {port_class_desc}') from e

            logger.error('failed to initialize port from driver %s: %s', port_class_desc, e, exc_info=True)

    # Map IDs
    for port in ports:
        if await port.get_attr('virtual'):
            continue

        old_id = port.get_id()
        new_id = settings.port_mappings.get(old_id)
        if not new_id:
            continue

        if new_id in _ports_by_id:
            if raise_on_error:
                raise PortLoadError(f'Cannot map port {old_id} to {new_id}: new id already exists')

            logger.error('cannot map port %s to %s: new id already exists', old_id, new_id)
            continue

        port = _ports_by_id.get(old_id)
        if not port:
            if raise_on_error:
                raise PortLoadError(f'Cannot map port {old_id} to {new_id}: no such port')

            logger.error('cannot map port %s to %s: no such port', old_id, new_id)
            continue

        try:
            port.map_id(new_id)

        except Exception as e:
            if raise_on_error:
                raise PortLoadError(f'Cannot map port {old_id} to {new_id}') from e

            port.error('cannot map to %s: %s', new_id, e)

        _ports_by_id.pop(old_id)
        _ports_by_id[port.get_id()] = port

    # Load created ports
    for port in ports:
        try:
            await port.load()

        except Exception as e:
            if raise_on_error:
                raise PortLoadError(f'Failed to load {port}') from e

            port.error('failed to load: %s', port, e, exc_info=True)

        await port.trigger_add()

    return ports


async def load_one(cls: Union[str, type], args: Dict[str, Any]) -> BasePort:
    port_args = [dict(driver=cls, **args)]
    ports = await load(port_args, raise_on_error=True)

    return ports[0]


def get(port_id: str) -> Optional[BasePort]:
    return _ports_by_id.get(port_id)


def get_all() -> List[BasePort]:
    return list(_ports_by_id.values())


async def init() -> None:
    # Use raise_on_error=False because we prefer a partial successful startup rather than a failed one
    await load(settings.ports, raise_on_error=False)


async def cleanup() -> None:

    async def cleanup_port(port: BasePort) -> None:
        await port.disable()
        await port.cleanup()

    tasks = [cleanup_port(port) for port in _ports_by_id.values()]
    if tasks:
        await asyncio.wait(tasks)


async def reset() -> None:
    logger.debug('clearing ports persisted data')
    await persist.remove(BasePort.PERSIST_COLLECTION)




import logging
import operator

from typing import Any, Dict, Iterable, List, Optional, Set, Tuple

import redis

from qtoggleserver.core.typing import GenericJSONDict
from qtoggleserver.persist import BaseDriver
from qtoggleserver.persist.typing import Id, Record
from qtoggleserver.utils import json as json_utils


FILTER_OP_MAPPING = {
    'gt': operator.gt,
    'ge': operator.ge,
    'lt': operator.lt,
    'le': operator.le,
    'in': lambda a, b: a in b
}

logger = logging.getLogger(__name__)


class DuplicateRecordId(redis.RedisError):
    pass


class RedisDriver(BaseDriver):
    def __init__(
        self,
        host: str = '127.0.0.1',
        port: int = 6379,
        db: int = 0,
        history_support: bool = False,
        **kwargs
    ) -> None:
        logger.debug('connecting to %s:%s/%s', host, port, db)

        self._client: redis.Redis = redis.StrictRedis(
            host=host,
            port=port,
            db=db,
            encoding='utf8',
            decode_responses=True
        )

        self._history_support: bool = history_support

    async def query(
        self,
        collection: str,
        fields: Optional[List[str]],
        filt: Dict[str, Any],
        sort: List[Tuple[str, bool]],
        limit: Optional[int]
    ) -> Iterable[Record]:

        db_records = []

        if isinstance(filt.get('id'), Id):  # Look for specific record id
            filt = dict(filt)
            id_ = filt.pop('id')
            db_record = self._client.hgetall(self._make_record_key(collection, id_))

            # Apply filter criteria
            if db_record and self._filter_matches(db_record, filt):
                db_record['id'] = id_
                db_records.append(db_record)

        else:  # No single specific id in filt
            # Look through all records from this collection, iterating through set
            for id_ in self._client.sscan_iter(self._make_set_key(collection)):
                # Retrieve the db record
                db_record = self._client.hgetall(self._make_record_key(collection, id_))
                db_record['id'] = id_

                # Apply filter criteria
                if self._filter_matches(db_record, filt):
                    db_records.append(db_record)

        # Sort
        for field, rev in reversed(sort):
            if field == 'id':
                db_records.sort(key=lambda r: int(r['id']), reverse=rev)

            else:
                db_records.sort(key=lambda r: self._value_from_db(r.get(field)), reverse=rev)

        # Apply limit
        if limit is not None:
            db_records = db_records[:limit]

        # Transform from db record and return
        if fields is not None:
            fields = set(fields)

        return (self._record_from_db(dbr, fields) for dbr in db_records)

    async def insert(self, collection: str, record: Record) -> Id:
        # Make sure we have an id
        record = dict(record)
        id_ = record.pop('id', None)
        if id_ is None:
            id_ = self._get_next_id(collection)

        key = self._make_record_key(collection, id_)
        set_key = self._make_set_key(collection)

        # Check for duplicates
        if self._client.sismember(set_key, id_):
            raise DuplicateRecordId(id_)

        # Adapt the record to db
        db_record = self._record_to_db(record)

        # Actually insert the record, but only if it's not empty
        if db_record:
            self._client.hset(key, mapping=db_record)

        # Add the id to set
        self._client.sadd(set_key, id_)

        return id_

    async def update(self, collection: str, record_part: Record, filt: Dict[str, Any]) -> int:
        # Adapt the record part to db
        db_record_part = self._record_to_db(record_part)

        modified_count = 0

        if isinstance(filt.get('id'), Id):
            filt = dict(filt)
            id_ = filt.pop('id')
            key = self._make_record_key(collection, id_)

            # Retrieve the db record
            db_record = self._client.hgetall(key)
            if db_record and self._filter_matches(db_record, filt):
                self._client.hset(key, mapping=db_record_part)
                modified_count = 1

        else:  # No single specific id in filt
            # Look through all records from this collection, iterating through set
            for id_ in self._client.sscan_iter(self._make_set_key(collection)):
                key = self._make_record_key(collection, id_)

                # Retrieve the db record
                db_record = self._client.hgetall(key)
                db_record['id'] = id_

                # Apply filter criteria
                if not self._filter_matches(db_record, filt):
                    continue

                # Actually update the record
                if db_record_part:
                    self._client.hset(key, mapping=db_record_part)

                else:
                    self._client.delete(key)

                modified_count += 1

        return modified_count

    async def replace(self, collection: str, id_: Id, record: Record) -> bool:
        # Adapt the record to db
        new_db_record = self._record_to_db(record)
        new_db_record.pop('id', None)  # Never add the id together with other fields

        skey = self._make_set_key(collection)
        key = self._make_record_key(collection, id_)

        if not self._client.sismember(skey, id_):
            return False  # No record found, no replacing

        # Remove existing record
        self._client.delete(key)

        # Insert the new record, if not empty
        if new_db_record:
            self._client.hset(key, mapping=new_db_record)

        # Make sure the id is present in set
        self._client.sadd(skey, id_)

        return True

    async def remove(self, collection: str, filt: Dict[str, Any]) -> int:
        removed_count = 0

        if isinstance(filt.get('id'), Id):
            filt = dict(filt)
            id_ = filt.pop('id')
            key = self._make_record_key(collection, id_)
            db_record = self._client.hgetall(key)

            # Actually remove the record
            if db_record and self._filter_matches(db_record, filt):
                self._client.delete(key)
                removed_count = 1

            # Remove the id from set
            self._client.srem(self._make_set_key(collection), id_)

        else:  # No single specific id in filt
            ids_to_remove = set()

            # Look through all records from this collection, iterating through set
            for id_ in self._client.sscan_iter(self._make_set_key(collection)):
                key = self._make_record_key(collection, id_)

                # Retrieve the db record
                db_record = self._client.hgetall(key)
                db_record['id'] = id_

                # Apply filter criteria
                if not self._filter_matches(db_record, filt):
                    continue

                # Actually remove the record
                self._client.delete(key)

                # Remember ids to remove from set
                ids_to_remove.add(id_)

                removed_count += 1

            # Remove the ids from set
            for id_ in ids_to_remove:
                self._client.srem(self._make_set_key(collection), id_)

        return removed_count

    async def cleanup(self) -> None:
        logger.debug('disconnecting redis client')

        self._client.close()

    def is_history_supported(self) -> bool:
        return self._history_support

    def _filter_matches(self, db_record: GenericJSONDict, filt: Dict[str, Any]) -> bool:
        for key, value in filt.items():
            try:
                db_record_value = db_record[key]

            except KeyError:
                return False

            if key == 'id':
                record_value = db_record_value

            else:
                record_value = self._value_from_db(db_record_value)

            if not self._filter_value_matches(record_value, value):
                return False

        return True

    @staticmethod
    def _filter_value_matches(record_value: Any, filt_value: Any) -> bool:
        if isinstance(filt_value, dict):  # filter with operators
            for op, v in filt_value.items():
                op_func = FILTER_OP_MAPPING[op]
                if not op_func(record_value, v):
                    return False

            return True

        else:  # Assuming simple value
            return record_value == filt_value

    def _get_next_id(self, collection: str) -> Id:
        return str(self._client.incr(self._make_sequence_key(collection)))

    @classmethod
    def _record_from_db(cls, db_record: GenericJSONDict, fields: Optional[Set[str]] = None) -> Record:
        if fields is not None:
            return {k: (cls._value_from_db(v) if k != 'id' else v) for k, v in db_record.items() if k in fields}

        else:
            return {k: (cls._value_from_db(v) if k != 'id' else v) for k, v in db_record.items()}

    @classmethod
    def _record_to_db(cls, record: Record) -> GenericJSONDict:
        return {k: (cls._value_to_db(v) if k != 'id' else v) for k, v in record.items()}

    @staticmethod
    def _value_to_db(value: Any) -> str:
        return json_utils.dumps(value, allow_extended_types=True)

    @staticmethod
    def _value_from_db(value: str) -> Any:
        return json_utils.loads(value, allow_extended_types=True)

    @staticmethod
    def _make_record_key(collection: str, id_: Id) -> str:
        if id_:
            return f'{collection}:{id_}'

        else:
            return collection

    @staticmethod
    def _make_set_key(collection: str) -> str:
        return f'{collection}-id-set'

    @staticmethod
    def _make_sequence_key(collection: str) -> str:
        return f'{collection}-id-sequence'




import logging
import re

from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

import bson
import pymongo.database
import pymongo.errors

from qtoggleserver.persist import BaseDriver
from qtoggleserver.persist.typing import Id, Record


logger = logging.getLogger(__name__)

_OBJECT_ID_RE = re.compile('^[0-9a-f]{24}$')
DEFAULT_DB = 'qtoggleserver'

FILTER_OP_MAPPING = {
    'gt': '$gt',
    'ge': '$gte',
    'lt': '$lt',
    'le': '$lte',
    'in': '$in'
}


class MongoDriver(BaseDriver):
    def __init__(self, host: str = '127.0.0.1', port: str = 27017, db: str = DEFAULT_DB, **kwargs) -> None:
        logger.debug('connecting to %s:%s/%s', host, port, db)

        self._client: pymongo.MongoClient = pymongo.MongoClient(
            host,
            port,
            serverSelectionTimeoutMS=200,
            connectTimeoutMS=200
        )
        self._db: pymongo.database.Database = self._client[db]

    async def query(
        self,
        collection: str,
        fields: Optional[List[str]],
        filt: Dict[str, Any],
        sort: List[Tuple[str, bool]],
        limit: Optional[int]
    ) -> Iterable[Record]:

        if fields:
            fields = dict((f, 1) for f in fields)
            if 'id' in fields:
                fields['_id'] = 1

            else:
                fields['_id'] = 0

        if 'id' in filt:
            filt = dict(filt)
            filt['_id'] = self._id_to_db_rec(filt.pop('id'))

        db_filt = self._filt_to_db(filt)

        q = self._db[collection].find(db_filt, fields)

        if len(sort) > 0:
            sort = [(f, [pymongo.ASCENDING, pymongo.DESCENDING][r]) for f, r in sort]
            q = q.sort(sort)

        if limit is not None:
            q = q.limit(limit)

        return self._query_gen_wrapper(q)

    async def insert(self, collection: str, record: Record) -> Id:
        record = dict(record)
        if 'id' in record:
            record['_id'] = self._id_to_db(record.pop('id'))

        return self._id_from_db(self._db[collection].insert_one(record).inserted_id)

    async def update(self, collection: str, record_part: Record, filt: Dict[str, Any]) -> int:
        if 'id' in record_part:
            record_part = dict(record_part)
            record_part['_id'] = self._id_to_db(record_part.pop('id'))

        if 'id' in filt:
            filt = dict(filt)
            filt['_id'] = self._id_to_db_rec(filt.pop('id'))

        db_filt = self._filt_to_db(filt)

        return self._db[collection].update_many(db_filt, {'$set': record_part}, upsert=False).modified_count

    async def replace(self, collection: str, id_: Id, record: Record) -> bool:
        record = dict(record)
        id_ = self._id_to_db(id_)
        record['_id'] = id_

        matched = self._db[collection].replace_one({'_id': id_}, record, upsert=False).matched_count

        return matched > 0

    async def remove(self, collection: str, filt: Dict[str, Any]) -> int:
        if 'id' in filt:
            filt = dict(filt)
            filt['_id'] = self._id_to_db_rec(filt.pop('id'))

        db_filt = self._filt_to_db(filt)

        return self._db[collection].delete_many(db_filt).deleted_count

    async def ensure_index(self, collection: str, index: List[Tuple[str, bool]]) -> None:
        index = [(f, [pymongo.ASCENDING, pymongo.DESCENDING][r]) for f, r in index]

        try:
            self._db[collection].create_index(index)

        except pymongo.errors.DuplicateKeyError:
            logger.debug('index already exists')

    async def cleanup(self) -> None:
        logger.debug('disconnecting mongo client')

        self._client.close()

    def is_history_supported(self) -> bool:
        return True

    @classmethod
    def _query_gen_wrapper(cls, q: Iterable[Record]) -> Iterable[Record]:
        for r in q:
            if '_id' in r:
                r['id'] = cls._id_from_db(r.pop('_id'))

            yield r

    @staticmethod
    def _id_to_db(id_: str) -> Union[str, bson.ObjectId]:
        if isinstance(id_, str) and _OBJECT_ID_RE.match(id_):
            return bson.ObjectId(id_)

        else:
            return id_

    @staticmethod
    def _id_from_db(id_: Union[str, bson.ObjectId]) -> str:
        if isinstance(id_, bson.ObjectId):
            return str(id_)

        else:
            return id_

    def _id_to_db_rec(self, obj: Union[Dict, List, str]) -> Union[Dict, List, bson.ObjectId]:
        if isinstance(obj, dict):
            return {key: self._id_to_db_rec(value) for key, value in obj.items()}

        elif isinstance(obj, list):
            return [self._id_to_db(value) for value in obj]

        else:  # Assuming str
            return self._id_to_db(obj)

    @staticmethod
    def _filt_to_db(filt: Dict[str, Any]) -> Dict[str, Any]:
        db_filt = {}
        for key, value in filt.items():
            if isinstance(value, dict):  # filter with operators
                value = {FILTER_OP_MAPPING[k]: v for k, v in value.items()}

            db_filt[key] = value

        return db_filt




import time

from .base import Evaluated
from .functions import function, Function


@function('TIME')
class TimeFunction(Function):
    MIN_ARGS = MAX_ARGS = 0
    DEPS = ['second']

    async def eval(self) -> Evaluated:
        return int(time.time())


@function('TIMEMS')
class TimeMSFunction(Function):
    MIN_ARGS = MAX_ARGS = 0
    DEPS = ['millisecond']

    async def eval(self) -> Evaluated:
        return int(time.time() * 1000)




import hashlib
import re


def generate_password(master_key: str, mac_address: str, username: str) -> str:
    mac_stripped = re.sub('[^A-F0-9]', '', mac_address.upper())
    to_hash = f'{master_key}:{mac_stripped}:{username}'

    return hashlib.sha256(to_hash.encode()).hexdigest()[:32]  # Max 32 chars allowed in password




import logging
import threading

from typing import Any, Dict, Iterable, List, Optional, Union

from qtoggleserver.conf import settings
from qtoggleserver.utils import conf as conf_utils
from qtoggleserver.utils import dynload as dynload_utils
from qtoggleserver.utils import json as json_utils

from .base import BaseDriver
from .typing import Id, Record


logger = logging.getLogger(__name__)

_thread_local: threading.local = threading.local()


def _get_driver() -> BaseDriver:
    if not hasattr(_thread_local, 'driver'):
        driver_args = conf_utils.obj_to_dict(settings.persist)
        driver_class_path = driver_args.pop('driver')

        try:
            logger.debug('loading persistence driver %s', driver_class_path)
            driver_class = dynload_utils.load_attr(driver_class_path)
            _thread_local.driver = driver_class(**driver_args)

        except Exception as e:
            logger.error('failed to load persistence driver %s: %s', driver_class_path, e, exc_info=True)

            raise

    return _thread_local.driver


def is_history_supported() -> bool:
    return _get_driver().is_history_supported()


async def query(
    collection: str,
    fields: Optional[List[str]] = None,
    filt: Optional[Dict[str, Any]] = None,
    sort: Optional[Union[str, List[str]]] = None,
    limit: Optional[int] = None
) -> Iterable[Record]:

    if logger.getEffectiveLevel() <= logging.DEBUG:
        logger.debug(
            'querying %s (%s) where %s (sort=%s, limit=%s)',
            collection,
            json_utils.dumps(fields) if fields else 'all fields',
            json_utils.dumps(filt, allow_extended_types=True),
            json_utils.dumps(sort),
            json_utils.dumps(limit),
        )

    filt = filt or {}
    sort = sort or []
    if isinstance(sort, str):
        sort = [sort]

    # Transform '-field' into (field, reverse)
    sort = [
        (s[1:], True) if s.startswith('-') else (s, False)
        for s in sort
    ]

    return await _get_driver().query(collection, fields, filt, sort, limit)


async def get(collection: str, id_: Id) -> Optional[Record]:
    logger.debug('getting record with id %s from %s', id_, collection)

    records = list(await _get_driver().query(collection, fields=None, filt={'id': id_}, sort=[], limit=1))
    if len(records) > 1:
        logger.warning('more than one record with same id %s found in collection %s', id_, collection)

    if records:
        return records[0]

    return None


async def get_value(name: str, default: Optional[Any] = None) -> Any:
    logger.debug('getting value of %s', name)

    records = list(await _get_driver().query(name, fields=None, filt={}, sort=[], limit=2))
    if len(records) > 1:
        logger.warning('more than one record found in single-value collection %s', name)

        record = records[0]

    elif len(records) > 0:
        record = records[0]

    else:
        return default

    return record['value']


async def set_value(name: str, value: Any) -> None:
    if logger.getEffectiveLevel() <= logging.DEBUG:
        logger.debug('setting %s to %s', name, json_utils.dumps(value, allow_extended_types=True))

    driver = _get_driver()
    record = {'value': value}

    records = list(await driver.query(name, fields=['id'], filt={}, sort=[], limit=2))
    if len(records) > 1:
        logger.warning('more than one record found in single-value collection %s', name)

        id_ = records[0]['id']

    elif len(records) > 0:
        id_ = records[0]['id']

    else:
        id_ = None

    if id_ is None:
        await driver.insert(name, record)

    else:
        await driver.replace(name, id_, record)


async def insert(collection: str, record: Record) -> Id:
    if logger.getEffectiveLevel() <= logging.DEBUG:
        logger.debug('inserting %s into %s', json_utils.dumps(record, allow_extended_types=True), collection)

    return await _get_driver().insert(collection, record)


async def update(collection: str, record_part: Record, filt: Optional[Dict[str, Any]] = None) -> int:
    if logger.getEffectiveLevel() <= logging.DEBUG:
        logger.debug(
            'updating %s where %s with %s',
            collection,
            json_utils.dumps(filt or {}, allow_extended_types=True),
            json_utils.dumps(record_part, allow_extended_types=True)
        )

    count = await _get_driver().update(collection, record_part, filt or {})

    logger.debug('modified %s records in %s', count, collection)

    return count


async def replace(collection: str, id_: Id, record: Record) -> bool:
    if logger.getEffectiveLevel() <= logging.DEBUG:
        logger.debug(
            'replacing record with id %s with %s in %s',
            id_,
            json_utils.dumps(record, allow_extended_types=True),
            collection
        )

    record = dict(record, id=id_)  # Make sure the new record contains the id field
    replaced = await _get_driver().replace(collection, id_, record)
    if replaced:
        logger.debug('replaced record with id %s in %s', id_, collection)

        return False

    else:
        await _get_driver().insert(collection, dict(record, id=id_))
        logger.debug('inserted record with id %s in %s', id_, collection)

        return True


async def remove(collection: str, filt: Optional[Dict[str, Any]] = None) -> int:
    if logger.getEffectiveLevel() <= logging.DEBUG:
        logger.debug('removing from %s where %s', collection, json_utils.dumps(filt or {}, allow_extended_types=True))

    count = await _get_driver().remove(collection, filt or {})

    logger.debug('removed %s records from %s', count, collection)

    return count


async def ensure_index(collection: str, index: Union[str, List[str]]) -> None:
    if isinstance(index, str):
        index = [index]

    # Transform '-field' into (field, reverse)
    index = [
        (i[1:], True) if i.startswith('-') else (i, False)
        for i in index
    ]

    if logger.getEffectiveLevel() <= logging.DEBUG:
        logger.debug('ensuring index %s in %s', json_utils.dumps(index), collection)

    await _get_driver().ensure_index(collection, index)


async def cleanup() -> None:
    logger.debug('cleaning up')

    return await _get_driver().cleanup()




import asyncio
import copy
import datetime
import hashlib
import logging
import re
import socket
import sys
import time

from typing import Optional

from qtoggleserver import system
from qtoggleserver import version
from qtoggleserver.conf import settings
from qtoggleserver.core.typing import Attributes, AttributeDefinitions, GenericJSONDict
from qtoggleserver.utils import json as json_utils
from qtoggleserver.utils.cmd import run_set_cmd

from . import events as device_events


logger = logging.getLogger(__name__)

ATTRDEFS = {
    'name': {
        'type': 'string',
        'modifiable': True,
        'persisted': True,
        'min': 1,
        'max': 32,
        'pattern': r'^[_a-zA-Z][_a-zA-Z0-9-]{0,31}$',
        'standard': True
    },
    'display_name': {
        'type': 'string',
        'modifiable': True,
        'max': 64,
        'standard': True
    },
    'version': {
        'type': 'string',
        'standard': True
    },
    'api_version': {
        'type': 'string',
        'standard': True
    },
    'vendor': {
        'type': 'string',
        'standard': True
    },
    'admin_password': {
        'type': 'string',
        'modifiable': True,
        'max': 32,
        'standard': True
    },
    'normal_password': {
        'type': 'string',
        'modifiable': True,
        'max': 32,
        'standard': True
    },
    'viewonly_password': {
        'type': 'string',
        'modifiable': True,
        'max': 32,
        'standard': True
    },
    'flags': {
        'type': ['string'],
        'standard': True
    },
    'virtual_ports': {
        'type': 'number',
        'enabled': lambda: bool(settings.core.virtual_ports),
        'standard': True
    },
    'uptime': {
        'type': 'number',
        'standard': True
    },
    'date': {
        'type': 'number',
        'modifiable': lambda: system.date.has_set_date_support(),
        'persisted': False,
        'standard': False  # Having standard False here enables exposing of definition (needed for non-modifiable)
    },
    'timezone': {
        'type': 'string',
        'modifiable': True,
        'persisted': False,
        'choices': [{'value': zone} for zone in system.date.get_timezones()],
        'enabled': lambda: system.date.has_timezone_support(),
        'standard': False  # Having standard False here enables exposing of definition (needed for choices)
    },
    'wifi_ssid': {
        'type': 'string',
        'max': 32,
        'modifiable': True,
        'persisted': False,
        'enabled': lambda: system.net.has_wifi_support(),
        'standard': True
    },
    'wifi_key': {
        'type': 'string',
        'max': 64,
        'modifiable': True,
        'persisted': False,
        'enabled': lambda: system.net.has_wifi_support(),
        'standard': True
    },
    'wifi_bssid': {
        'type': 'string',
        'pattern': r'^([a-fA-F0-9]{2}:[a-fA-F0-9]{2}:[a-fA-F0-9]{2}:[a-fA-F0-9]{2}:[a-fA-F0-9]{2}:[a-fA-F0-9]{2})?$',
        'modifiable': True,
        'persisted': False,
        'enabled': lambda: system.net.has_wifi_support(),
        'standard': True
    },
    'wifi_bssid_current': {
        'type': 'string',
        'modifiable': False,
        'persisted': False,
        'enabled': lambda: system.net.has_wifi_support(),
        'standard': True
    },
    'wifi_signal_strength': {
        'type': 'number',
        'modifiable': False,
        'persisted': False,
        'enabled': lambda: system.net.has_wifi_support(),
        'standard': True
    },
    'ip_address': {
        'type': 'string',
        'pattern': r'^(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})?$',
        'modifiable': True,
        'persisted': False,
        'enabled': lambda: system.net.has_ip_support(),
        'standard': True
    },
    'ip_netmask': {
        'type': 'number',
        'min': 0,
        'max': 31,
        'integer': True,
        'modifiable': True,
        'persisted': False,
        'enabled': lambda: system.net.has_ip_support(),
        'standard': True
    },
    'ip_gateway': {
        'type': 'string',
        'pattern': r'^(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})?$',
        'modifiable': True,
        'persisted': False,
        'enabled': lambda: system.net.has_ip_support(),
        'standard': True
    },
    'ip_dns': {
        'type': 'string',
        'pattern': r'^(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})?$',
        'modifiable': True,
        'persisted': False,
        'enabled': lambda: system.net.has_ip_support(),
        'standard': True
    },
    'ip_address_current': {
        'type': 'string',
        'modifiable': False,
        'persisted': False,
        'enabled': lambda: system.net.has_ip_support(),
        'standard': True
    },
    'ip_netmask_current': {
        'type': 'number',
        'modifiable': False,
        'persisted': False,
        'enabled': lambda: system.net.has_ip_support(),
        'standard': True
    },
    'ip_gateway_current': {
        'type': 'string',
        'modifiable': False,
        'persisted': False,
        'enabled': lambda: system.net.has_ip_support(),
        'standard': True
    },
    'ip_dns_current': {
        'type': 'string',
        'modifiable': False,
        'persisted': False,
        'enabled': lambda: system.net.has_ip_support(),
        'standard': True
    },
    'cpu_usage': {
        'type': 'number',
        'min': 0,
        'max': 100,
        'modifiable': False,
        'persisted': False,
        'standard': True
    },
    'mem_usage': {
        'type': 'number',
        'min': 0,
        'max': 100,
        'modifiable': False,
        'persisted': False,
        'standard': True
    },
    'storage_usage': {
        'type': 'number',
        'min': 0,
        'max': 100,
        'modifiable': False,
        'persisted': False,
        'enabled': lambda: system.storage.has_storage_support(),
        'standard': True
    },
    'temperature': {
        'type': 'number',
        'modifiable': False,
        'persisted': False,
        'enabled': lambda: system.temperature.has_temperature_support(),
        'min': lambda: settings.system.temperature.min,
        'max': lambda: settings.system.temperature.max,
        'standard': False  # Having standard False here enables exposing of definition (needed for min/max)
    },
    'battery_level': {
        'type': 'number',
        'min': 0,
        'max': 100,
        'modifiable': False,
        'persisted': False,
        'enabled': lambda: system.battery.has_battery_support(),
        'standard': True
    }
}

EMPTY_PASSWORD_HASH = hashlib.sha256(b'').hexdigest()

WIFI_RSSI_EXCELLENT = -50
WIFI_RSSI_GOOD = -60
WIFI_RSSI_FAIR = -70

NETWORK_ATTRS_WATCH_INTERVAL = 10


name: str = re.sub(r'[^a-zA-Z0-9_-]', '', socket.gethostname())
if not re.match('^[a-zA-Z_]', name):  # Make sure name starts with a letter or underscore
    name = f'q{name}'
name = name[:32]

display_name: str = ''
admin_password_hash: Optional[str] = None
normal_password_hash: Optional[str] = None
viewonly_password_hash: Optional[str] = None

_schema: Optional[GenericJSONDict] = None
_attrdefs: Optional[AttributeDefinitions] = None
_attrs_watch_task: Optional[asyncio.Task] = None


class DeviceAttributeError(Exception):
    def __init__(self, error: str, attribute: str) -> None:
        self.error: str = error
        self.attribute: str = attribute


def get_attrdefs() -> AttributeDefinitions:
    global _attrdefs

    if _attrdefs is None:
        logger.debug('initializing attribute definitions')
        _attrdefs = copy.deepcopy(ATTRDEFS)

        # Transform all callable values into corresponding results
        for n, attrdef in _attrdefs.items():
            for k, v in attrdef.items():
                if callable(v):
                    attrdef[k] = v()

    return _attrdefs


def get_schema(loose: bool = False) -> GenericJSONDict:
    global _schema

    # Use cached value, but only when loose is false, as loose schema is never cached
    if _schema is not None and not loose:
        return _schema

    schema = {
        'type': 'object',
        'properties': {},
        'additionalProperties': loose
    }

    for n, attrdef in get_attrdefs().items():
        if not attrdef.get('modifiable'):
            continue

        attr_schema = dict(attrdef)

        enabled = attr_schema.pop('enabled', True)
        if not enabled:
            continue

        if attr_schema['type'] == 'string':
            if 'min' in attr_schema:
                attr_schema['minLength'] = attr_schema.pop('min')

            if 'max' in attr_schema:
                attr_schema['maxLength'] = attr_schema.pop('max')

        elif attr_schema['type'] == 'number':
            if attr_schema.get('integer'):
                attr_schema['type'] = 'integer'

            if 'min' in attr_schema:
                attr_schema['minimum'] = attr_schema.pop('min')

            if 'max' in attr_schema:
                attr_schema['maximum'] = attr_schema.pop('max')

        if 'choices' in attrdef:
            attr_schema['enum'] = [c['value'] for c in attr_schema.pop('choices')]

        attr_schema.pop('persisted', None)
        attr_schema.pop('modifiable', None)
        attr_schema.pop('standard', None)

        schema['properties'][n] = attr_schema

    if not loose:
        _schema = schema

    return schema


def get_attrs() -> Attributes:
    from qtoggleserver.core import api as core_api
    from qtoggleserver.core import history as core_history

    attrs = {
        'name': name,
        'display_name': display_name,
        'version': version.VERSION,
        'api_version': core_api.API_VERSION,
        'vendor': version.VENDOR,
        'uptime': system.uptime(),

        # Never disclose passwords
        'admin_password': '' if admin_password_hash == EMPTY_PASSWORD_HASH else 'set',
        'normal_password': '' if normal_password_hash == EMPTY_PASSWORD_HASH else 'set',
        'viewonly_password': '' if viewonly_password_hash == EMPTY_PASSWORD_HASH else 'set'
    }

    flags = ['expressions']
    if settings.system.fwupdate.driver:
        flags.append('firmware')

    if settings.core.backup_support:
        flags.append('backup')

    if core_history.is_enabled():
        flags.append('history')

    if settings.core.listen_support:
        flags.append('listen')

    if settings.slaves.enabled:
        flags.append('master')

    if settings.reverse.enabled:
        flags.append('reverse')

    if settings.core.sequences_support:
        flags.append('sequences')

    if settings.core.ssl_support:
        flags.append('ssl')

    if settings.webhooks.enabled:
        flags.append('webhooks')

    attrs['flags'] = flags

    if settings.core.virtual_ports:
        attrs['virtual_ports'] = settings.core.virtual_ports

    if system.date.has_real_date_time():
        attrs['date'] = int(time.time())

    if system.date.has_timezone_support():
        attrs['timezone'] = system.date.get_timezone()

    if system.net.has_wifi_support():
        wifi_config = system.net.get_wifi_config()
        attrs['wifi_ssid'] = wifi_config['ssid']
        attrs['wifi_key'] = wifi_config['psk']
        attrs['wifi_bssid'] = wifi_config['bssid']

        if wifi_config['bssid_current']:
            attrs['wifi_bssid_current'] = wifi_config['bssid_current']

        rssi = wifi_config['rssi_current']
        if rssi:
            rssi = int(rssi)
            if rssi >= WIFI_RSSI_EXCELLENT:
                strength = 3

            elif rssi >= WIFI_RSSI_GOOD:
                strength = 2

            elif rssi >= WIFI_RSSI_FAIR:
                strength = 1

            else:
                strength = 0

            attrs['wifi_signal_strength'] = strength

    if system.net.has_ip_support():
        ip_config = system.net.get_ip_config()
        attrs['ip_address'] = ip_config['address']
        attrs['ip_netmask'] = int(ip_config['netmask'] or 0)
        attrs['ip_gateway'] = ip_config['gateway']
        attrs['ip_dns'] = ip_config['dns']

        if 'address_current' in ip_config:
            attrs['ip_address_current'] = ip_config['address_current']
        if 'netmask_current' in ip_config:
            attrs['ip_netmask_current'] = int(ip_config['netmask_current'] or 0)
        if 'gateway_current' in ip_config:
            attrs['ip_gateway_current'] = ip_config['gateway_current']
        if 'dns_current' in ip_config:
            attrs['ip_dns_current'] = ip_config['dns_current']

    attrs['cpu_usage'] = system.get_cpu_usage()
    attrs['mem_usage'] = system.get_mem_usage()

    if system.storage.has_storage_support():
        attrs['storage_usage'] = system.storage.get_storage_usage()

    if system.temperature.has_temperature_support():
        attrs['temperature'] = system.temperature.get_temperature()

    if system.battery.has_battery_support():
        attrs['battery_level'] = system.battery.get_battery_level()

    return attrs


def set_attrs(attrs: Attributes, ignore_extra: bool = False) -> bool:
    core_device_attrs = sys.modules[__name__]

    reboot_required = False
    attrdefs = get_attrdefs()

    wifi_attrs = {}
    ip_attrs = {}

    for n, value in attrs.items():
        # A few attributes may carry sensitive information, so treat them separately and do not log their values
        if n.count('password') or n == 'wifi_key':
            logger.debug('setting device attribute %s', n)

        else:
            logger.debug('setting device attribute %s = %s', n, json_utils.dumps(value))

        try:
            attrdef = attrdefs[n]

        except KeyError:
            if ignore_extra:
                continue

            else:
                raise

        if not attrdef.get('modifiable'):
            if not ignore_extra:
                raise DeviceAttributeError('attribute-not-modifiable', n)

        # Treat passwords separately, as they are not persisted as given, but hashed first
        if n.endswith('_password') and hasattr(core_device_attrs, f'{n}_hash'):
            # Call password set command, if available
            if settings.core.passwords.set_cmd:
                run_set_cmd(
                    settings.core.passwords.set_cmd,
                    cmd_name='password',
                    log_values=False,
                    username=n[:-9],
                    password=value
                )

            value = hashlib.sha256(value.encode()).hexdigest()
            n += '_hash'

            setattr(core_device_attrs, n, value)
            continue

        elif n.endswith('_password_hash') and hasattr(core_device_attrs, n):
            # FIXME: Password set command cannot be called with hash and we don't have clear-text password here.
            #        A solution would be to use sha256 crypt algorithm w/o salt for Unix password (watch for the special
            #        alphabet and for number of rounds defaulting to 5000)
            setattr(core_device_attrs, n, value)
            continue

        persisted = attrdef.get('persisted', attrdef.get('modifiable'))
        if persisted:
            setattr(core_device_attrs, n, value)

        if n == 'name' and settings.core.device_name.set_cmd:
            run_set_cmd(settings.core.device_name.set_cmd, cmd_name='device name', name=value)

        elif n == 'date' and system.date.has_set_date_support():
            date = datetime.datetime.utcfromtimestamp(value)
            system.date.set_date(date)

        elif n == 'timezone' and system.date.has_timezone_support():
            system.date.set_timezone(value)

        elif n in ('wifi_ssid', 'wifi_key', 'wifi_bssid') and system.net.has_wifi_support():
            k = n[5:]
            k = {
                'key': 'psk'
            }.get(k, k)
            wifi_attrs[k] = value

        elif n in ('ip_address', 'ip_netmask', 'ip_gateway', 'ip_dns') and system.net.has_ip_support():
            k = n[3:]
            ip_attrs[k] = value

    if wifi_attrs:
        wifi_config = system.net.get_wifi_config()

        for k, v in wifi_attrs.items():
            wifi_config[k] = v
            wifi_config = {k: v for k, v in wifi_config.items() if not k.endswith('_current')}

        system.net.set_wifi_config(**wifi_config)
        reboot_required = True

    if ip_attrs:
        ip_config = system.net.get_ip_config()

        for k, v in ip_attrs.items():
            ip_config[k] = v
            ip_config = {k: v for k, v in ip_config.items() if not k.endswith('_current')}
            ip_config['netmask'] = str(ip_config['netmask'])

        system.net.set_ip_config(**ip_config)
        reboot_required = True

    return reboot_required


def to_json() -> GenericJSONDict:
    attrdefs = copy.deepcopy(get_attrdefs())
    filtered_attrdefs = {}
    for n, attrdef in attrdefs.items():
        if attrdef.pop('standard', False):
            continue

        enabled = attrdef.pop('enabled', True)
        if not enabled:
            continue

        attrdef.pop('persisted', None)
        attrdef.pop('pattern', None)

        filtered_attrdefs[n] = attrdef

    result = dict(get_attrs())
    result['definitions'] = filtered_attrdefs

    return result


def _check_net_data_changed(data: dict) -> bool:
    changed = False

    if system.net.has_wifi_support():
        wifi_config = system.net.get_wifi_config()
        old_wifi_config = data.get('wifi_config')
        if old_wifi_config != wifi_config:
            data['wifi_config'] = wifi_config
            changed = True

    if system.net.has_ip_support():
        ip_config = system.net.get_ip_config()
        old_ip_config = data.get('ip_config')
        if old_ip_config != ip_config:
            data['ip_config'] = ip_config
            changed = True

    return changed


async def _attrs_watch_loop() -> None:
    last_net_time = time.time()
    last_net_data = {}

    try:
        while True:
            now = time.time()
            changed = False
            if now - last_net_time >= NETWORK_ATTRS_WATCH_INTERVAL:
                try:
                    if _check_net_data_changed(last_net_data):
                        logger.debug('network attributes data changed')
                        changed = True

                except Exception as e:
                    logger.error('network attributes data check failed: %s', e, exc_info=True)

                last_net_time = now

            if changed:
                await device_events.trigger_update()

            await asyncio.sleep(1)

    except asyncio.CancelledError:
        logger.debug('attributes watch task cancelled')


async def init() -> None:
    global _attrs_watch_task

    logger.debug('starting attributes watch task')
    _attrs_watch_task = asyncio.create_task(_attrs_watch_loop())


async def cleanup() -> None:
    logger.debug('stopping attributes watch task')
    if _attrs_watch_task:
        _attrs_watch_task.cancel()
        await _attrs_watch_task




from __future__ import annotations

import asyncio
import datetime
import json
import logging
import time

from typing import Dict, List, Optional, Tuple

from tornado import httpclient

from qtoggleserver.conf import settings
from qtoggleserver.core.device import attrs as core_device_attrs
from qtoggleserver.core.typing import Attributes, GenericJSONDict
from qtoggleserver.slaves import utils as salves_utils
from qtoggleserver.system import ap
from qtoggleserver.system import dhcp
from qtoggleserver.system import dns
from qtoggleserver.system import net
from qtoggleserver.utils import asyncio as asyncio_utils
from qtoggleserver.utils import cmd as cmd_utils

from .exceptions import DiscoverException


_PATH_PREFIXES = ['', '/api']
_INTERFACE_CACHE_TIMEOUT = 60

logger = logging.getLogger(__name__)

_discover_task: Optional[asyncio.Task] = None
_discovered_devices: Optional[Dict[str, DiscoveredDevice]] = None
_finish_timer: Optional[asyncio_utils.Timer] = None

# TODO: replace these with a common cache service with integrated timeout management
_interface: Optional[Tuple[str]] = None
_interface_time: float = 0


class DiscoveredDevice:
    def __init__(
        self,
        ap_client: ap.APClient,
        scheme: str,
        port: int,
        path: str,
        admin_password: str,
        attrs: Attributes
    ) -> None:

        self.ap_client: ap.APClient = ap_client
        self.scheme: str = scheme
        self.port: int = port
        self.path: str = path
        self.admin_password: str = admin_password
        self.attrs: Attributes = attrs

    def __str__(self) -> str:
        return f'discovered AP device {self.attrs["name"]} at {self.ap_client.ip_address}'

    def to_json(self) -> GenericJSONDict:
        return {
            'mac_address': self.ap_client.mac_address,
            'ip_address': self.ap_client.ip_address,
            'timestamp': datetime.datetime.timestamp(self.ap_client.moment),
            'scheme': self.scheme,
            'hostname': self.ap_client.hostname,
            'port': self.port,
            'path': self.path,
            'admin_password': self.admin_password,
            'attrs': self.attrs
        }


def get_interface() -> Optional[str]:
    global _interface
    global _interface_time

    ap_settings = settings.slaves.discover.ap
    if ap_settings.interface:
        return ap_settings.interface

    if ap_settings.interface_cmd:
        now = time.time()
        if now - _interface_time > _INTERFACE_CACHE_TIMEOUT:
            _interface = None

        if _interface is not None:
            return _interface[0]

        result = cmd_utils.run_get_cmd(
            ap_settings.interface_cmd,
            cmd_name='AP client discover interface',
            exc_class=DiscoverException,
            required_fields=['interface']
        )

        _interface = (result.get('interface') or None,)
        _interface_time = now

        return _interface[0]


async def discover(timeout: int) -> None:
    global _discover_task
    global _discovered_devices
    global _finish_timer

    # If discover task is already running, just await it instead of starting a new one
    if _discover_task:
        await _discover_task
        return

    _discover_task = asyncio.create_task(_discover(timeout))

    if _finish_timer:
        _finish_timer.cancel()

    _finish_timer = asyncio_utils.Timer(settings.slaves.discover.ap.finish_timeout, _finish_timeout)

    devices_list = await _discover_task

    logger.debug('discovered %d devices', len(devices_list))

    _discovered_devices = {d.attrs['name']: d for d in devices_list}


async def finish() -> None:
    global _discover_task
    global _discovered_devices
    global _finish_timer

    logger.debug('finishing')

    if _discover_task:
        _discover_task.cancel()
        try:
            await _discover_task

        except Exception as e:
            logger.error('discover task error: %s', e, exc_info=True)

        _discover_task = None

    if ap.is_running():
        await ap.stop()

    if _finish_timer:
        _finish_timer.cancel()
        _finish_timer = None

    _discovered_devices = None


def get_discovered_devices() -> Optional[Dict[str, DiscoveredDevice]]:
    return _discovered_devices


async def configure(discovered_device: DiscoveredDevice, attrs: Attributes) -> DiscoveredDevice:
    ap_client = discovered_device.ap_client

    # Wi-Fi settings of the real (target) network
    if 'wifi_ssid' not in attrs:
        logger.warning('no SSID/PSK available to configure device')

    if 'admin_password' not in attrs:
        logger.debug('generating password for %s', discovered_device)
        discovered_device.admin_password = salves_utils.generate_password(
            core_device_attrs.admin_password_hash,
            discovered_device.ap_client.mac_address,
            'admin'
        )

        attrs['admin_password'] = discovered_device.admin_password

        # Also set normal and view-only password, if exposed
        if 'normal_password' in discovered_device.attrs:
            attrs['normal_password'] = salves_utils.generate_password(
                core_device_attrs.admin_password_hash,
                discovered_device.ap_client.mac_address,
                'normal'
            )

        if 'viewonly_password' in discovered_device.attrs:
            attrs['viewonly_password'] = salves_utils.generate_password(
                core_device_attrs.admin_password_hash,
                discovered_device.ap_client.mac_address,
                'viewonly'
            )

    else:
        logger.debug('using supplied password for %s', discovered_device)
        discovered_device.admin_password = attrs['admin_password']

    network_configured = attrs.get('wifi_ssid') is not None
    if network_configured:
        # Find client's future IP address first
        dhcp_interface = settings.slaves.discover.dhcp_interface or net.get_default_interface()
        if not dhcp_interface:
            raise DiscoverException('No DHCP interface')

        try:
            reply = await dhcp.request(
                interface=dhcp_interface,
                timeout=settings.slaves.discover.dhcp_timeout,
                mac_address=ap_client.mac_address,
                hostname=ap_client.hostname
            )

        except dhcp.DHCPTimeout:
            logger.warning('could not determine future device IP address of %s', discovered_device)
            reply = None

        if reply:
            adjusted_ap_client = ap.APClient(
                mac_address=ap_client.mac_address,
                ip_address=reply.ip_address,
                hostname=ap_client.hostname,
                moment=ap_client.moment
            )

            discovered_device = DiscoveredDevice(
                ap_client=adjusted_ap_client,
                scheme=discovered_device.scheme,
                port=discovered_device.port,
                path=discovered_device.path,
                admin_password=discovered_device.admin_password,
                attrs=discovered_device.attrs
            )

            dns.set_custom_dns_mapping(ap_client.hostname, reply.ip_address, timeout=60)

    logger.debug('configuring %s', discovered_device)
    await ap_client.request('PATCH', f'{discovered_device.path}/device', body=attrs)
    logger.debug('%s successfully configured', discovered_device)

    # Remove configured device from discovered list
    _discovered_devices.pop(discovered_device.attrs['name'])

    if network_configured:
        logger.debug('waiting for %s to connect to new network', discovered_device)
        await asyncio.sleep(5)  # Device requires at least 5 seconds to connect to new network
        start_time = time.time()
        while True:
            try:
                await discovered_device.ap_client.request(
                    'GET',
                    '/device',
                    no_log=True,
                    admin_password=discovered_device.admin_password
                )
                logger.debug('%s connected to new network', discovered_device)
                break

            except Exception:
                if time.time() - start_time > settings.slaves.long_timeout:
                    logger.error('timeout waiting for %s to connect to new network', discovered_device)
                    break

                await asyncio.sleep(1)

    return discovered_device


async def _discover(timeout: int) -> List[DiscoveredDevice]:
    logger.debug('starting discovery')

    # (Re)start our AP
    ap_settings = settings.slaves.discover.ap

    if ap.is_running():
        await ap.stop()

    ap.start(
        interface=get_interface(),
        ssid=ap_settings.ssid,
        psk=ap_settings.psk,
        own_ip=ap_settings.own_ip,
        mask_len=ap_settings.mask_len,
        start_ip=ap_settings.start_ip,
        stop_ip=ap_settings.stop_ip,
        hostapd_binary=ap_settings.hostapd_binary,
        hostapd_cli_binary=ap_settings.hostapd_cli_binary,
        dnsmasq_binary=ap_settings.dnsmasq_binary,
        hostapd_log=ap_settings.hostapd_log,
        dnsmasq_log=ap_settings.dnsmasq_log
    )

    # Wait for clients to connect to our AP
    try:
        for _ in range(timeout):
            await asyncio.sleep(1)
            await ap.update()
            if not ap.is_running():
                raise DiscoverException('AP stopped unexpectedly')

        clients = ap.get_clients()

    except asyncio.CancelledError:
        logger.debug('discover task cancelled')
        clients = []

    for client in clients:
        logger.debug(
            'discovered AP client "%s" with IP %s and MAC %s',
            client.hostname,
            client.ip_address,
            client.mac_address
        )

    if not clients:
        logger.debug('no clients discovered')

    # See if we're really dealing with qToggle devices
    discovered_devices = []
    for client in clients:
        try:
            discovered_devices.append(await _query_client(client))

        except Exception as e:
            logger.error('client query failed: %s', e, exc_info=True)

    return discovered_devices


async def _finish_timeout() -> None:
    global _finish_timer

    logger.debug('finish timeout')
    _finish_timer = None
    await finish()


async def _query_client(ap_client: ap.APClient) -> Optional[DiscoveredDevice]:
    logger.debug('querying %s', ap_client)

    for prefix in _PATH_PREFIXES:
        try:
            attrs = await ap_client.request('GET', f'{prefix}/device')
            break

        except (httpclient.HTTPError, json.JSONDecodeError):
            continue

    else:
        raise DiscoverException('Could not find device API endpoint')

    if not isinstance(attrs, dict):
        logger.error('%s has returned invalid device attributes response', ap_client)
        return None

    for field in ['name', 'vendor', 'version']:
        if field not in attrs:
            logger.error('%s has returned invalid device attributes response', ap_client)
            return None

    return DiscoveredDevice(
        ap_client=ap_client,
        scheme='http',  # TODO: add support for HTTPS schemes
        port=80,
        path=prefix,
        admin_password='',
        attrs=attrs
    )


async def init() -> None:
    pass


async def cleanup() -> None:
    await finish()




from typing import Any

from qtoggleserver.core import api as core_api
from qtoggleserver.core import events as core_events
from qtoggleserver.core.typing import GenericJSONDict


# We can't use proper type annotations for slaves in this module because that would create unsolvable circular imports.
# Therefore we use "Any" type annotation for Slave instances.


class SlaveDeviceEvent(core_events.Event):
    def __init__(self, slave: Any, timestamp: float = None) -> None:
        self._slave = slave

        super().__init__(timestamp)

    def __str__(self) -> str:
        return f'{self._type}({self._slave.get_name()}) event'

    def get_slave(self) -> Any:
        return self._slave


class SlaveDeviceAdd(SlaveDeviceEvent):
    REQUIRED_ACCESS = core_api.ACCESS_LEVEL_ADMIN
    TYPE = 'slave-device-add'

    async def get_params(self) -> GenericJSONDict:
        return self.get_slave().to_json()


class SlaveDeviceRemove(SlaveDeviceEvent):
    REQUIRED_ACCESS = core_api.ACCESS_LEVEL_ADMIN
    TYPE = 'slave-device-remove'

    async def get_params(self) -> GenericJSONDict:
        return {'name': self.get_slave().get_name()}


class SlaveDeviceUpdate(SlaveDeviceEvent):
    REQUIRED_ACCESS = core_api.ACCESS_LEVEL_ADMIN
    TYPE = 'slave-device-update'

    async def get_params(self) -> GenericJSONDict:
        return self.get_slave().to_json()

    def is_duplicate(self, event: core_events.Event) -> bool:
        return isinstance(event, self.__class__) and event.get_slave() == self.get_slave()




import datetime
import json
import math

from typing import Any, Union

import jsonpointer


JSON_CONTENT_TYPE = 'application/json; charset=utf-8'

TYPE_FIELD = '__t'
VALUE_FIELD = '__v'
DATE_FORMAT = '%Y-%m-%d'
DATETIME_FORMAT = '%Y-%m-%dT%H:%M:%S.%fZ'
DATE_TYPE = '__d'
DATETIME_TYPE = '__dt'


def _replace_nan_inf_rec(obj: Any, replace_value: Any) -> Any:
    if isinstance(obj, dict):
        new_obj = {}
        for k, v in obj.items():
            if isinstance(v, float) and (math.isnan(v) or math.isinf(v)):
                new_obj[k] = replace_value

            else:
                new_obj[k] = _replace_nan_inf_rec(v, replace_value)

    elif isinstance(obj, (list, tuple, set)):
        new_obj = []
        for e in obj:
            if isinstance(e, float) and (math.isnan(e) or math.isinf(e)):
                new_obj.append(replace_value)

            else:
                new_obj.append(_replace_nan_inf_rec(e, replace_value))

    else:
        new_obj = obj

    return new_obj


def _resolve_refs_rec(obj: Any, root_obj: Any) -> Any:
    if isinstance(obj, dict):
        if len(obj.keys()) == 1 and list(obj.keys())[0] == '$ref':
            ref = list(obj.values())[0]
            ref = ref[1:]  # Skip starting hash
            return jsonpointer.resolve_pointer(root_obj, ref)

        for k, v in obj.items():
            obj[k] = _resolve_refs_rec(v, root_obj)

    elif isinstance(obj, list):
        for i, e in enumerate(obj):
            obj[i] = _resolve_refs_rec(e, root_obj)

    return obj


def encode_default_json(obj: Any) -> Any:
    if isinstance(obj, datetime.datetime):
        return {
            TYPE_FIELD: DATETIME_TYPE,
            VALUE_FIELD: obj.strftime(DATETIME_FORMAT)
        }

    elif isinstance(obj, datetime.date):
        return {
            TYPE_FIELD: DATE_TYPE,
            VALUE_FIELD: obj.strftime(DATE_FORMAT)
        }

    elif isinstance(obj, (set, tuple)):
        return list(obj)

    else:
        raise TypeError()


def decode_json_hook(obj: dict) -> Any:
    __t = obj.get(TYPE_FIELD)
    if __t is not None:
        __v = obj.get(VALUE_FIELD)
        if __t == DATE_TYPE:
            try:
                return datetime.datetime.strptime(__v, DATE_FORMAT).date()

            except ValueError:
                pass

        elif __t == DATETIME_TYPE:
            try:
                return datetime.datetime.strptime(__v, DATETIME_FORMAT)

            except ValueError:
                pass

    return obj


def dumps(obj: Any, allow_extended_types: bool = False, **kwargs) -> str:
    # Treat primitive types separately to gain just a bit of performance
    if isinstance(obj, str):
        return '"' + obj + '"'

    elif isinstance(obj, bool):
        return ['false', 'true'][obj]

    elif isinstance(obj, (int, float)):
        if math.isinf(obj) or math.isnan(obj):
            return 'null'

        return str(obj)

    elif obj is None:
        return 'null'

    else:
        if allow_extended_types:
            return json.dumps(obj, default=encode_default_json, allow_nan=True, **kwargs)

        else:
            try:
                return json.dumps(obj, allow_nan=False, **kwargs)

            except ValueError:
                # Retry again by replacing Infinity and NaN values with None
                obj = _replace_nan_inf_rec(obj, replace_value=None)
                return json.dumps(obj, allow_nan=False, **kwargs)


def loads(s: Union[str, bytes], resolve_refs: bool = False, allow_extended_types: bool = False, **kwargs) -> Any:
    object_hook = decode_json_hook if allow_extended_types else None
    obj = json.loads(s, object_hook=object_hook, **kwargs)

    if resolve_refs:
        obj = _resolve_refs_rec(obj, root_obj=obj)

    return obj




import asyncio
import logging
import psutil
import re
import subprocess
import tempfile
import time

from typing import Optional, TextIO

from .exceptions import APException


BINARY = 'hostapd'
CLI_BINARY = 'hostapd_cli'

CONF_TEMPLATE = (
    'ssid={ssid}\n'
    'wpa_passphrase={psk}\n'
    'wpa=2\n'
    'wpa_key_mgmt=WPA-PSK\n'
    'interface={interface}\n'
    'channel=1\n'
    'driver=nl80211\n'
    'ctrl_interface=/var/run/hostapd\n'
)

CONF_NO_PSK_TEMPLATE = (
    'ssid={ssid}\n'
    'interface={interface}\n'
    'channel=1\n'
    'driver=nl80211\n'
    'ctrl_interface=/var/run/hostapd\n'
)

STOP_TIMEOUT = 2


logger = logging.getLogger(__name__)


class HostAPDException(APException):
    pass


class HostAPD:
    def __init__(
        self,
        ssid: str,
        psk: Optional[str],
        interface: str,
        hostapd_binary: Optional[str] = None,
        hostapd_cli_binary: Optional[str] = None,
        hostapd_log: Optional[str] = None
    ) -> None:

        self._ssid: str = ssid
        self._psk: Optional[str] = psk
        self._interface: str = interface
        self._binary: Optional[str] = hostapd_binary
        self._cli_binary: Optional[str] = hostapd_cli_binary
        self._log: Optional[str] = hostapd_log

        self._conf_file: Optional[TextIO] = None
        self._log_file: Optional[TextIO] = None
        self._process: Optional[subprocess.Popen] = None

    def is_alive(self) -> bool:
        return (self._process is not None) and (self._process.poll() is None)

    def is_running(self) -> bool:
        return self._process is not None

    def start(self) -> None:
        logger.debug(
            'starting hostapd with SSID "%s" and PSK %s',
            self._ssid,
            f'"{self._psk}"' if self._psk else 'empty'
        )

        binary = self._binary or self._find_binary(BINARY)
        if not binary:
            raise HostAPDException('Could not find %s binary', BINARY)

        conf_template = CONF_TEMPLATE if self._psk else CONF_NO_PSK_TEMPLATE
        conf = conf_template.format(ssid=self._ssid, psk=self._psk, interface=self._interface)

        self._log_file = open(self._log, 'wt')
        self._conf_file = tempfile.NamedTemporaryFile(mode='wt')
        self._conf_file.write(conf)
        self._conf_file.flush()

        self._process = subprocess.Popen(
            [binary, self._conf_file.name],
            stdout=self._log_file,
            stderr=subprocess.STDOUT
        )

    async def stop(self) -> None:
        logger.debug('stopping hostapd')

        if self._process:
            self._disassociate_clients()
            await asyncio.sleep(1)  # Allow 1 second for hostapd to honor the command

            self._process.terminate()

            begin_time = time.time()
            while self._process.poll() is None:
                await asyncio.sleep(0.1)
                if time.time() - begin_time > STOP_TIMEOUT:
                    break

            # If process could not be stopped in time, kill it
            if self._process.poll() is None:
                logger.error('failed to stop hostapd within %d seconds, killing it', STOP_TIMEOUT)
                self._process.kill()
                await asyncio.sleep(1)
                self._process.poll()  # We want no zombies

            self._process = None

        if self._conf_file:
            self._conf_file.close()
            self._conf_file = None

        if self._log_file:
            self._log_file.close()
            self._log_file = None

    def _disassociate_clients(self) -> None:
        cli_binary = self._cli_binary or self._find_binary(CLI_BINARY)
        if not cli_binary:
            raise HostAPDException('Could not find %s binary', CLI_BINARY)

        try:
            output = subprocess.check_output([cli_binary, 'list'], stderr=subprocess.DEVNULL).decode().strip()

        except subprocess.CalledProcessError:
            logger.error('command hostapd_cli list failed')
            return

        lines = output.split('\n')
        mac_addresses = [line for line in lines if re.match(r'^[a-f0-9:]{17}$', line)]

        for mac_address in mac_addresses:
            logger.debug('disassociating client with MAC %s', mac_address.upper())
            try:
                subprocess.check_call(
                    [cli_binary, 'disassociate', mac_address],
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL
                )

            except subprocess.CalledProcessError:
                logger.error('command hostapd_cli disassociate failed')

    def _find_binary(self, binary: str) -> Optional[str]:
        try:
            return subprocess.check_output(['which', binary], stderr=subprocess.DEVNULL).decode().strip()

        except subprocess.CalledProcessError:
            return None

    @staticmethod
    def is_already_running() -> bool:
        return BINARY in (p.name() for p in psutil.process_iter())




import logging

from qtoggleserver.conf import settings
from qtoggleserver.core.api.funcs import backup as backup_api_funcs
from qtoggleserver.core.api.funcs import device as device_api_funcs
from qtoggleserver.core.api.funcs import firmware as firmware_api_funcs
from qtoggleserver.core.api.funcs import ports as ports_api_funcs
from qtoggleserver.core.api.funcs import reverse as reverse_api_funcs
from qtoggleserver.core.api.funcs import various as various_api_funcs
from qtoggleserver.core.api.funcs import webhooks as webhooks_api_funcs
from qtoggleserver.frontend.api import funcs as frontend_api_funcs
from qtoggleserver.slaves.api import funcs as slaves_api_funcs
from qtoggleserver.system.api import funcs as system_api_funcs

from .base import APIHandler, BaseHandler, NoSuchFunction


logger = logging.getLogger(__name__)


class NoSuchFunctionHandler(BaseHandler):
    pass


class DeviceHandler(APIHandler):
    async def get(self) -> None:
        await self.call_api_func(device_api_funcs.get_device)

    async def put(self) -> None:
        await self.call_api_func(device_api_funcs.put_device, default_status=204)

    async def patch(self) -> None:
        await self.call_api_func(device_api_funcs.patch_device, default_status=204)


class ResetHandler(APIHandler):
    async def post(self) -> None:
        await self.call_api_func(various_api_funcs.post_reset, default_status=204)


class FirmwareHandler(APIHandler):
    async def get(self) -> None:
        await self.call_api_func(firmware_api_funcs.get_firmware)

    async def patch(self) -> None:
        await self.call_api_func(firmware_api_funcs.patch_firmware, default_status=204)


class BackupEndpointsHandler(APIHandler):
    async def get(self) -> None:
        await self.call_api_func(backup_api_funcs.get_backup_endpoints)


class AccessHandler(APIHandler):
    async def get(self) -> None:
        await self.call_api_func(various_api_funcs.get_access)


class SlaveDevicesHandler(APIHandler):
    async def get(self) -> None:
        await self.call_api_func(slaves_api_funcs.get_slave_devices)

    async def put(self) -> None:
        await self.call_api_func(slaves_api_funcs.put_slave_devices, default_status=204)

    async def post(self) -> None:
        await self.call_api_func(slaves_api_funcs.post_slave_devices, default_status=201)


class SlaveDeviceHandler(APIHandler):
    async def patch(self, name: str) -> None:
        await self.call_api_func(slaves_api_funcs.patch_slave_device, name=name, default_status=204)

    async def delete(self, name: str) -> None:
        await self.call_api_func(slaves_api_funcs.delete_slave_device, name=name, default_status=204)


class SlaveDeviceEventsHandler(APIHandler):
    AUTH_ENABLED = False  # We'll take care of the authentication inside API call functions

    async def post(self, name: str) -> None:
        await self.call_api_func(slaves_api_funcs.post_slave_device_events, name=name, default_status=204)


class SlaveDeviceForwardHandler(APIHandler):
    async def get(self, name: str, path: str) -> None:
        await self.call_api_func(slaves_api_funcs.slave_device_forward, name=name, path=path)

    post = patch = put = delete = get


class DiscoveredHandler(APIHandler):
    async def get(self) -> None:
        await self.call_api_func(slaves_api_funcs.get_discovered)

    async def delete(self) -> None:
        await self.call_api_func(slaves_api_funcs.delete_discovered, default_status=204)


class DiscoveredDeviceHandler(APIHandler):
    async def patch(self, name: str) -> None:
        await self.call_api_func(slaves_api_funcs.patch_discovered_device, name=name)


class PortsHandler(APIHandler):
    async def get(self) -> None:
        await self.call_api_func(ports_api_funcs.get_ports)

    async def put(self) -> None:
        await self.call_api_func(ports_api_funcs.put_ports, default_status=204)

    async def post(self) -> None:
        if not settings.core.virtual_ports:
            raise NoSuchFunction()

        await self.call_api_func(ports_api_funcs.post_ports, default_status=201)


class PortHandler(APIHandler):
    async def delete(self, port_id: str) -> None:
        if not settings.core.virtual_ports:
            raise NoSuchFunction()

        await self.call_api_func(ports_api_funcs.delete_port, port_id=port_id, default_status=204)

    async def patch(self, port_id: str) -> None:
        await self.call_api_func(ports_api_funcs.patch_port, port_id=port_id, default_status=204)


class PortValueHandler(APIHandler):
    async def get(self, port_id: str) -> None:
        await self.call_api_func(ports_api_funcs.get_port_value, port_id=port_id)

    async def patch(self, port_id: str) -> None:
        await self.call_api_func(ports_api_funcs.patch_port_value, port_id=port_id, default_status=204)


class PortSequenceHandler(APIHandler):
    async def patch(self, port_id: str) -> None:
        await self.call_api_func(ports_api_funcs.patch_port_sequence, port_id=port_id, default_status=204)


class PortHistoryHandler(APIHandler):
    async def get(self, port_id: str) -> None:
        await self.call_api_func(ports_api_funcs.get_port_history, port_id=port_id)

    async def delete(self, port_id: str) -> None:
        await self.call_api_func(ports_api_funcs.delete_port_history, port_id=port_id, default_status=204)


class WebhooksHandler(APIHandler):
    async def get(self) -> None:
        await self.call_api_func(webhooks_api_funcs.get_webhooks)

    async def put(self) -> None:
        await self.call_api_func(webhooks_api_funcs.put_webhooks, default_status=204)


class ListenHandler(APIHandler):
    async def get(self) -> None:
        await self.call_api_func(various_api_funcs.get_listen)


class ReverseHandler(APIHandler):
    async def get(self) -> None:
        await self.call_api_func(reverse_api_funcs.get_reverse)

    async def put(self) -> None:
        await self.call_api_func(reverse_api_funcs.put_reverse, default_status=204)


class DashboardPanelsHandler(APIHandler):
    async def get(self) -> None:
        await self.call_api_func(frontend_api_funcs.get_panels)

    async def put(self) -> None:
        await self.call_api_func(frontend_api_funcs.put_panels, default_status=204)


class PrefsHandler(APIHandler):
    async def get(self) -> None:
        await self.call_api_func(frontend_api_funcs.get_prefs)

    async def put(self) -> None:
        await self.call_api_func(frontend_api_funcs.put_prefs, default_status=204)


class FrontendHandler(APIHandler):
    async def get(self) -> None:
        await self.call_api_func(frontend_api_funcs.get_frontend)

    async def put(self) -> None:
        await self.call_api_func(frontend_api_funcs.put_frontend, default_status=204)


class SystemHandler(APIHandler):
    async def get(self) -> None:
        await self.call_api_func(system_api_funcs.get_system)

    async def put(self) -> None:
        await self.call_api_func(system_api_funcs.put_system, default_status=204)


class IntrospectHandler(APIHandler):
    async def post(self) -> None:
        await self.call_api_func(various_api_funcs.post_introspect)




from typing import Optional

import psutil

from qtoggleserver.conf import settings


def has_storage_support() -> bool:
    return bool(settings.system.storage.path)


def get_storage_usage() -> Optional[int]:
    if not settings.system.storage.path:
        return None

    return int(psutil.disk_usage(settings.system.storage.path).percent)




from qtoggleserver.core import api as core_api
from qtoggleserver.core.device import attrs as core_device_attrs
from qtoggleserver.core.typing import Attributes, GenericJSONDict

from .base import Event


class DeviceEvent(Event):
    def __init__(self, timestamp: float = None) -> None:
        self._attrs: Attributes = core_device_attrs.to_json()

        super().__init__(timestamp)

    def get_attrs(self) -> Attributes:
        return self._attrs


class DeviceUpdate(DeviceEvent):
    REQUIRED_ACCESS = core_api.ACCESS_LEVEL_ADMIN
    TYPE = 'device-update'

    async def get_params(self) -> GenericJSONDict:
        return self.get_attrs()

    def is_duplicate(self, event: Event) -> bool:
        return isinstance(event, self.__class__)


class FullUpdate(DeviceEvent):
    REQUIRED_ACCESS = core_api.ACCESS_LEVEL_VIEWONLY
    TYPE = 'full-update'

    def is_duplicate(self, event: Event) -> bool:
        return isinstance(event, self.__class__)




import abc
import re

from typing import Optional, Set

from .base import Expression, Evaluated
from .exceptions import UnknownPortId, DisabledPort, PortValueUnavailable, UnexpectedCharacter

# Import core.ports after defining Expression, because core.ports.BasePort depends on Expression.
from qtoggleserver.core import ports as core_ports


class PortExpression(Expression, metaclass=abc.ABCMeta):
    def __init__(self, port_id: str) -> None:
        self.port_id: str = port_id

    def get_port(self) -> core_ports.BasePort:
        return core_ports.get(self.port_id)

    @staticmethod
    def parse(self_port_id: Optional[str], sexpression: str, pos: int) -> Expression:
        # Remove leading whitespace
        while sexpression and sexpression[0].isspace():
            sexpression = sexpression[1:]
            pos += 1

        # Remove trailing whitespace
        while sexpression and sexpression[-1].isspace():
            sexpression = sexpression[:-1]

        prefix = sexpression[0]
        port_id = sexpression[1:]

        if port_id:
            m = re.search(r'[^a-zA-Z0-9_.-]', port_id)
            if m:
                p = m.start()
                raise UnexpectedCharacter(port_id[p], p + pos + 2)

            if prefix == '$':
                return PortValue(port_id)

            else:  # Assuming prefix == '@'
                return PortRef(port_id)

        else:
            return SelfPortValue(self_port_id)


class PortValue(PortExpression):
    def __str__(self) -> str:
        return f'${self.port_id}'

    def get_deps(self) -> Set[str]:
        return {f'${self.port_id}'}

    async def eval(self) -> Evaluated:
        port = self.get_port()
        if not port:
            raise UnknownPortId(self.port_id)

        if not port.is_enabled():
            raise DisabledPort(self.port_id)

        value = port.get_last_read_value()
        if value is None:
            raise PortValueUnavailable(self.port_id)

        return value


class SelfPortValue(PortValue):
    def __str__(self) -> str:
        return '$'


class PortRef(PortExpression):
    def __str__(self) -> str:
        return f'@{self.port_id}'

    async def eval(self) -> Evaluated:
        port = self.get_port()
        if not port:
            raise UnknownPortId(self.port_id)

        return port




import logging

from typing import Dict, Optional

from qtoggleserver import persist
from qtoggleserver.core import ports as core_ports
from qtoggleserver.core.typing import GenericJSONDict, GenericJSONList, NullablePortValue, PortValue, PortValueChoices


logger = logging.getLogger(__name__)

_vport_args: Dict[str, GenericJSONDict] = {}


class VirtualPort(core_ports.Port):
    WRITABLE = True
    VIRTUAL = True

    def __init__(
        self,
        id_: str,
        type_: str,
        min_: Optional[float],
        max_: Optional[float],
        integer: Optional[bool],
        step: Optional[float],
        choices: Optional[PortValueChoices]
    ) -> None:

        super().__init__(id_)

        self._type: str = type_
        self._min: Optional[float] = min_
        self._max: Optional[float] = max_
        self._integer: Optional[bool] = integer
        self._step: Optional[float] = step
        self._choices: Optional[PortValueChoices] = choices

        self._value = self._virtual_value = None
        self._virtual_value: NullablePortValue = None

    def map_id(self, new_id: str) -> None:
        raise core_ports.PortError('Virtual ports cannot be mapped')

    async def read_value(self) -> NullablePortValue:
        return self._virtual_value

    async def write_value(self, value: PortValue) -> None:
        self._virtual_value = value


async def add(
    id_: str,
    type_: str,
    min_: Optional[float],
    max_: Optional[float],
    integer: Optional[bool],
    step: Optional[float],
    choices: Optional[PortValueChoices]
) -> None:

    settings = {
        'id': id_,
        'type': type_,
        'min': min_,
        'max': max_,
        'integer': integer,
        'step': step,
        'choices': choices
    }

    _vport_args[id_] = settings

    logger.debug('saving virtual port settings for %s', id_)
    await persist.replace('vports', id_, settings)


async def remove(port_id: str) -> None:
    _vport_args.pop(port_id, None)
    logger.debug('removing virtual port settings for %s', port_id)
    await persist.remove('vports', filt={'id': port_id})


def all_port_args() -> GenericJSONList:
    return [
        {'driver': VirtualPort, 'id_': port_id, **args}
        for port_id, args in _vport_args.items()
    ]


async def init() -> None:
    for entry in await persist.query('vports'):
        _vport_args[entry['id']] = {
            'type_': entry.get('type') or core_ports.TYPE_NUMBER,
            'min_': entry.get('min'),
            'max_': entry.get('max'),
            'integer': entry.get('integer'),
            'step': entry.get('step'),
            'choices': entry.get('choices')
        }

        logger.debug('loaded virtual port settings for %s', entry['id'])

    # Use raise_on_error=False because we prefer a partial successful startup rather than a failed one
    await core_ports.load(all_port_args(), raise_on_error=False)


async def reset() -> None:
    logger.debug('clearing virtual ports persisted data')
    await persist.remove('vports')




import abc
import asyncio
import logging
import os
import re

from typing import Optional

from qtoggleserver.lib import polled


W1_DEVICES_PATH = '/sys/bus/w1/devices'
SLAVE_FILE_NAME = 'w1_slave'


logger = logging.getLogger(__name__)


class OneWireException(Exception):
    pass


class OneWirePeripheralNotFound(OneWireException):
    def __init__(self, address: str) -> None:
        super().__init__(f'Peripheral @{address} not found')


class OneWireTimeout(OneWireException):
    def __init__(self, message: str = 'timeout') -> None:
        super().__init__(message)


class OneWirePeripheral(polled.PolledPeripheral, metaclass=abc.ABCMeta):
    logger = logger

    TIMEOUT = 5  # Seconds

    def __init__(self, *, address: str, **kwargs) -> None:
        super().__init__(**kwargs)

        self._address: str = address
        self._filename: Optional[str] = None
        self._data: Optional[str] = None

    def get_filename(self) -> str:
        if self._filename is None:
            self._filename = self._find_filename()

        return self._filename

    def _find_filename(self) -> str:
        address_parts = re.split('[^a-zA-Z0-9]', self._address)
        pat = address_parts[0] + '-0*' + ''.join(address_parts[1:])
        for name in os.listdir(W1_DEVICES_PATH):
            if re.match(pat, name, re.IGNORECASE):
                return os.path.join(W1_DEVICES_PATH, name, SLAVE_FILE_NAME)

        raise OneWirePeripheralNotFound(self._address)

    # def autodetect_addresses(self) -> List[str]:
    #     # TODO: make this method look only through specific device types (e.g. temperature sensors)
    #     # TODO: use this method in a more general peripheral autodetection routine
    #
    #     names = os.listdir(W1_DEVICES_PATH)
    #     names = [n for n in names if re.match('^[0-9]{2}-', n)]
    #     addresses = [re.sub('[^a-f0-9]', '', n, re.IGNORECASE) for n in names]
    #     addresses = [':'.join(a[2 * i: 2 * i + 2] for i in range(len(a) // 2)) for a in addresses]
    #
    #     return addresses

    def read(self) -> Optional[str]:
        data = self._data
        self._data = None

        return data

    def read_sync(self) -> Optional[str]:
        filename = self.get_filename()
        self.debug('opening file %s', filename)
        with open(filename, 'rt') as f:
            data = f.read()
            self.debug('read data: %s', data.replace('\n', '\\n'))

        return data

    async def poll(self) -> None:
        try:
            future = self.run_threaded(self.read_sync)
            self._data = await asyncio.wait_for(future, timeout=self.TIMEOUT)

        except asyncio.TimeoutError as e:
            raise OneWireTimeout('Timeout waiting for one-wire data from peripheral') from e

    async def handle_disable(self) -> None:
        await super().handle_disable()

        self._filename = None
        self._data = None


class OneWirePort(polled.PolledPort, metaclass=abc.ABCMeta):
    pass




from qtoggleserver.core import api as core_api
from qtoggleserver.core import events as core_events
from qtoggleserver.core.typing import GenericJSONDict, GenericJSONList


class FrontendEvent(core_events.Event):
    def __init__(self, request: core_api.APIRequest) -> None:
        self.request: core_api.APIRequest = request

        super().__init__()

    async def to_json(self) -> GenericJSONDict:
        result = await super().to_json()
        result['session_id'] = self.request.session_id

        return result

    def is_duplicate(self, event: core_events.Event) -> bool:
        return isinstance(event, self.__class__) and self.request.session_id == event.request.session_id


class DashboardUpdateEvent(FrontendEvent):
    REQUIRED_ACCESS = core_api.ACCESS_LEVEL_ADMIN
    TYPE = 'dashboard-update'

    def __init__(self, panels: GenericJSONList, **kwargs) -> None:
        self.panels: GenericJSONList = panels

        super().__init__(**kwargs)

    async def get_params(self) -> GenericJSONDict:
        return {
            'panels': self.panels
        }




import importlib
import logging

from typing import List, Optional

from qtoggleserver import persist
from qtoggleserver.conf import settings
from qtoggleserver.utils import json as json_utils
from qtoggleserver.utils.cmd import run_get_cmd

from . import attrs as device_attrs


logger = logging.getLogger(__name__)


def get_display_name() -> str:
    return device_attrs.display_name or device_attrs.name


async def load() -> None:
    data = await persist.get_value('device', {})

    # Attributes
    persisted_attrs = []
    for name, value in device_attrs.get_attrdefs().items():
        persisted = value.get('persisted', value.get('modifiable'))
        if callable(persisted):
            persisted = persisted()

        if persisted:
            persisted_attrs.append(name)

    for name in persisted_attrs:
        if name.endswith('password') and hasattr(device_attrs, name + '_hash'):
            name += '_hash'

        if name not in data:
            continue

        value = data[name]

        # A few attributes may carry sensitive information, so treat them separately and do not log their values
        if name.count('password') or name == 'wifi_key':
            logger.debug('loaded %s', name)

        else:
            logger.debug('loaded %s = %s', name, json_utils.dumps(value))

        setattr(device_attrs, name, value)

    # Device name
    if settings.core.device_name.get_cmd:
        result = run_get_cmd(settings.core.device_name.get_cmd, cmd_name='device name', required_fields=['name'])
        device_attrs.name = result['name']

    # Hash empty passwords
    if not device_attrs.admin_password_hash:
        device_attrs.admin_password_hash = device_attrs.EMPTY_PASSWORD_HASH
    if not device_attrs.normal_password_hash:
        device_attrs.normal_password_hash = device_attrs.EMPTY_PASSWORD_HASH
    if not device_attrs.viewonly_password_hash:
        device_attrs.viewonly_password_hash = device_attrs.EMPTY_PASSWORD_HASH


async def save() -> None:
    data = {}

    # Attributes
    persisted_attrs = []
    for name, value in device_attrs.get_attrdefs().items():
        persisted = value.get('persisted', value.get('modifiable'))
        if callable(persisted):
            persisted = persisted()

        if persisted:
            persisted_attrs.append(name)

    for name in persisted_attrs:
        if name.endswith('password') and hasattr(device_attrs, name + '_hash'):
            name += '_hash'

        data[name] = getattr(device_attrs, name)

    logger.debug('saving persisted data')
    await persist.set_value('device', data)


async def reset(preserve_attrs: Optional[List[str]] = None) -> None:
    preserve_attrs = preserve_attrs or []

    preserved_attrs = {}
    for name in preserve_attrs:
        preserved_attrs[name] = getattr(device_attrs, name, None)

    logger.debug('clearing persisted data')
    await persist.remove('device')
    importlib.reload(device_attrs)  # Reloads device attributes to default values

    for name, value in preserved_attrs.items():
        setattr(device_attrs, name, value)


async def init() -> None:
    logger.debug('loading persisted data')
    await load()

    logger.debug('initializing attributes')
    await device_attrs.init()


async def cleanup() -> None:
    logger.debug('cleaning up attributes')
    await device_attrs.cleanup()




import abc
import calendar
import datetime
import time

from qtoggleserver import system

from .base import Evaluated
from .exceptions import InvalidArgumentValue, EvalSkipped
from .functions import function, Function


class DateUnitFunction(Function, metaclass=abc.ABCMeta):
    MIN_ARGS = 0
    MAX_ARGS = 1
    DEPS = ['second']

    async def eval(self) -> Evaluated:
        if not system.date.has_real_date_time():
            raise EvalSkipped()

        if len(self.args) > 0:
            timestamp = int(await self.args[0].eval())

        else:
            timestamp = int(time.time())

        return self.extract_unit(datetime.datetime.fromtimestamp(timestamp))

    @abc.abstractmethod
    def extract_unit(self, dt: datetime.datetime) -> int:
        raise NotImplementedError()


@function('YEAR')
class YearFunction(DateUnitFunction):
    def extract_unit(self, dt: datetime.datetime) -> int:
        return dt.year


@function('MONTH')
class MonthFunction(DateUnitFunction):
    def extract_unit(self, dt: datetime.datetime) -> int:
        return dt.month


@function('DAY')
class DayFunction(DateUnitFunction):
    def extract_unit(self, dt: datetime.datetime) -> int:
        return dt.day


@function('DOW')
class DOWFunction(DateUnitFunction):
    def extract_unit(self, dt: datetime.datetime) -> int:
        return dt.weekday()


@function('LDOM')
class LDOMFunction(DateUnitFunction):
    def extract_unit(self, dt: datetime.datetime) -> int:
        return calendar.monthrange(dt.year, dt.month)[1]


@function('HOUR')
class HourFunction(DateUnitFunction):
    def extract_unit(self, dt: datetime.datetime) -> int:
        return dt.hour


@function('MINUTE')
class MinuteFunction(DateUnitFunction):
    def extract_unit(self, dt: datetime.datetime) -> int:
        return dt.minute


@function('SECOND')
class SecondFunction(DateUnitFunction):
    def extract_unit(self, dt: datetime.datetime) -> int:
        return dt.second


@function('MILLISECOND')
class MillisecondFunction(Function):
    MIN_ARGS = MAX_ARGS = 0
    DEPS = ['millisecond']

    async def eval(self) -> Evaluated:
        if not system.date.has_real_date_time():
            raise EvalSkipped()

        return datetime.datetime.now().microsecond // 1000


@function('DATE')
class DateFunction(Function):
    MIN_ARGS = MAX_ARGS = 6
    DEPS = ['second']
    UNIT_INDEX = {u: i + 1 for i, u in enumerate(('year', 'month', 'day', 'hour', 'minute', 'second'))}

    async def eval(self) -> Evaluated:
        if not system.date.has_real_date_time():
            raise EvalSkipped()

        eval_args = [int(await self.args[i].eval()) for i in range(self.MIN_ARGS)]

        try:
            return int(datetime.datetime(*eval_args).timestamp())

        except ValueError as e:
            unit = str(e).split()[0]
            index = self.UNIT_INDEX.get(unit)
            if index is None:
                raise

            raise InvalidArgumentValue(index, eval_args[index])


@function('BOY')
class BOYFunction(Function):
    MIN_ARGS = 0
    MAX_ARGS = 1
    DEPS = ['second']

    async def eval(self) -> Evaluated:
        if not system.date.has_real_date_time():
            raise EvalSkipped()

        now = datetime.datetime.now()

        n = 0
        if len(self.args) > 0:
            n = int(await self.args[0].eval())

        dt = datetime.datetime(now.year + n, 1, 1, 0, 0, 0)
        dt = dt.astimezone(tz=datetime.timezone.utc)

        return dt.timestamp()


@function('BOM')
class BOMFunction(Function):
    MIN_ARGS = 0
    MAX_ARGS = 1
    DEPS = ['second']

    async def eval(self) -> Evaluated:
        if not system.date.has_real_date_time():
            raise EvalSkipped()

        now = datetime.datetime.now()
        n = 0
        if len(self.args) > 0:
            n = int(await self.args[0].eval())

        year, month = now.year, now.month
        if n >= 0:
            for _ in range(n):
                if month < 12:
                    month += 1

                else:
                    year += 1
                    month = 1

        else:
            for _ in range(-n):
                if month > 1:
                    month -= 1

                else:
                    year -= 1
                    month = 12

        dt = datetime.datetime(year, month, 1, 0, 0, 0)
        dt = dt.astimezone(tz=datetime.timezone.utc)

        return dt.timestamp()


@function('BOW')
class BOWFunction(Function):
    MIN_ARGS = 0
    MAX_ARGS = 2
    DEPS = ['second']

    async def eval(self) -> Evaluated:
        if not system.date.has_real_date_time():
            raise EvalSkipped()

        n = 0
        s = 0
        if len(self.args) > 0:
            n = int(await self.args[0].eval())
            if len(self.args) > 1:
                s = int(await self.args[1].eval())

        now = datetime.datetime.now()
        dt = now.replace(hour=12)  # Using mid day practically avoids problems due to DST
        if s > 0:
            dt -= datetime.timedelta(days=dt.weekday() + 7 - s)

        else:
            dt -= datetime.timedelta(days=dt.weekday())

        year, month, day = dt.year, dt.month, dt.day
        if n >= 0:
            for _ in range(n):
                last_day = calendar.monthrange(year, month)[1]
                if day + 7 <= last_day:
                    day += 7

                else:
                    day = 7 - last_day + day
                    if month < 12:
                        month += 1

                    else:
                        year += 1
                        month = 1

        else:
            for _ in range(-n):
                if day > 7:
                    day -= 7

                else:
                    if month > 1:
                        month -= 1

                    else:
                        year -= 1
                        month = 12

                    last_day = calendar.monthrange(year, month)[1]
                    day = last_day - 7 + day

        dt = datetime.datetime(year, month, day)
        dt = dt.astimezone(tz=datetime.timezone.utc)

        return dt.timestamp()


@function('BOD')
class BODFunction(Function):
    MIN_ARGS = 0
    MAX_ARGS = 1
    DEPS = ['second']

    async def eval(self) -> Evaluated:
        if not system.date.has_real_date_time():
            raise EvalSkipped()

        now = datetime.datetime.now()
        n = 0
        if len(self.args) > 0:
            n = int(await self.args[0].eval())
        dt = now + datetime.timedelta(days=n)
        dt = dt.replace(hour=0, minute=0, second=0, microsecond=0)
        dt = dt.astimezone(tz=datetime.timezone.utc)

        return dt.timestamp()


@function('HMSINTERVAL')
class HMSIntervalFunction(Function):
    MIN_ARGS = MAX_ARGS = 6
    DEPS = ['second']

    async def eval(self) -> Evaluated:
        if not system.date.has_real_date_time():
            raise EvalSkipped()

        now = datetime.datetime.now().replace(microsecond=0)

        start_h, start_m, start_s, stop_h, stop_m, stop_s = await self.eval_args()

        if not (0 <= start_h <= 23):
            raise InvalidArgumentValue(1, start_h)

        if not (0 <= start_m <= 59):
            raise InvalidArgumentValue(2, start_m)

        if not (0 <= start_s <= 59):
            raise InvalidArgumentValue(3, start_s)

        if not (0 <= stop_h <= 23):
            raise InvalidArgumentValue(4, stop_h)

        if not (0 <= stop_m <= 59):
            raise InvalidArgumentValue(5, stop_m)

        if not (0 <= stop_s <= 59):
            raise InvalidArgumentValue(6, stop_s)

        start_time = datetime.time(int(start_h), int(start_m), int(start_s))
        stop_time = datetime.time(int(stop_h), int(stop_m), int(stop_s))

        start_dt = datetime.datetime.combine(now.date(), start_time)
        stop_dt = datetime.datetime.combine(now.date(), stop_time)

        return int(start_dt <= now <= stop_dt)


@function('MDINTERVAL')
class MDIntervalFunction(Function):
    MIN_ARGS = MAX_ARGS = 4
    DEPS = ['second']

    async def eval(self) -> Evaluated:
        if not system.date.has_real_date_time():
            raise EvalSkipped()

        now = datetime.datetime.now().replace(microsecond=0)

        start_m, start_d, stop_m, stop_d = await self.eval_args()

        if not (1 <= start_m <= 12):
            raise InvalidArgumentValue(1, start_m)

        try:
            start_dt = now.replace(month=int(start_m), day=int(start_d))

        except ValueError:
            raise InvalidArgumentValue(2, start_d)

        if not (1 <= stop_m <= 12):
            raise InvalidArgumentValue(3, stop_m)

        try:
            stop_dt = now.replace(month=int(stop_m), day=int(stop_d))

        except ValueError:
            raise InvalidArgumentValue(4, stop_d)

        return int(start_dt <= now <= stop_dt)




from __future__ import annotations

import abc
import logging

from typing import Optional, Tuple

from qtoggleserver.conf import settings
from qtoggleserver.utils import conf as conf_utils
from qtoggleserver.utils import dynload as dynload_utils


STATUS_IDLE = 'idle'
STATUS_CHECKING = 'checking'
STATUS_DOWNLOADING = 'downloading'
STATUS_EXTRACTING = 'extracting'
STATUS_VALIDATING = 'validating'
STATUS_FLASHING = 'flashing'
STATUS_RESTARTING = 'restarting'
STATUS_ERROR = 'error'


logger = logging.getLogger(__name__)

_driver: Optional[BaseDriver] = None


class FWUpdateException(Exception):
    pass


class FWUpdateDisabled(FWUpdateException):
    pass


class BaseDriver(metaclass=abc.ABCMeta):
    def __init__(self) -> None:
        pass

    @abc.abstractmethod
    async def get_current_version(self) -> str:
        pass

    @abc.abstractmethod
    async def get_latest(self) -> Tuple[str, str, str]:
        pass

    @abc.abstractmethod
    async def get_status(self) -> str:
        return STATUS_IDLE

    @abc.abstractmethod
    async def update_to_version(self, version: str) -> None:
        pass

    @abc.abstractmethod
    async def update_to_url(self, url: str) -> None:
        pass


def _get_fwupdate() -> BaseDriver:
    global _driver

    if not settings.system.fwupdate.driver:
        raise FWUpdateDisabled()

    if _driver is None:
        logger.debug('initializing fwupdate')

        driver_args = conf_utils.obj_to_dict(settings.system.fwupdate)
        driver_class_path = driver_args.pop('driver')

        driver_class = dynload_utils.load_attr(driver_class_path)
        _driver = driver_class(**driver_args)

    return _driver


async def get_current_version() -> str:
    current_version = await _get_fwupdate().get_current_version()
    logger.debug('current version: %s', current_version)

    return current_version


async def get_latest() -> Tuple[str, str, str]:
    latest_version, latest_date, latest_url = await _get_fwupdate().get_latest()
    logger.debug('latest version: %s/%s at %s', latest_version, latest_date, latest_url)

    return latest_version, latest_date, latest_url


async def get_status() -> str:
    status = await _get_fwupdate().get_status()
    logger.debug('status: %s', status)

    return status


async def update_to_version(version: str) -> None:
    logger.debug('updating to version %s', version)

    await _get_fwupdate().update_to_version(version)


async def update_to_url(url: str) -> None:
    logger.debug('updating to url %s', url)

    await _get_fwupdate().update_to_url(url)



#!/usr/bin/env python

import code

from qtoggleserver import commands


async def main() -> bool:
    import qtoggleserver  # noqa: F401; Required for locals()

    code.interact(local=locals())

    return False  # Don't run loop afterwards


def execute() -> None:
    commands.execute(main())


if __name__ == '__main__':
    execute()




import time

from typing import List, Optional, Set, Tuple

from . import TIME_JUMP_THRESHOLD

from .base import Evaluated
from .exceptions import EvalSkipped
from .functions import function, Function


@function('DELAY')
class DelayFunction(Function):
    MIN_ARGS = MAX_ARGS = 2
    DEPS = ['millisecond']
    HISTORY_SIZE = 1024

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._queue: List[Tuple[int, float]] = []
        self._last_value: Optional[float] = None
        self._current_value: Optional[float] = None

    async def eval(self) -> Evaluated:
        time_ms = int(time.time() * 1000)

        value, delay = await self.eval_args()

        if self._current_value is None:
            self._current_value = value

        # Detect value transitions and build history
        if value != self._last_value:
            self._last_value = value

            # Drop elements from queue if history size reached
            while len(self._queue) >= self.HISTORY_SIZE:
                self._queue.pop(0)

            self._queue.append((time_ms, value))

        # Process history
        while self._queue and (time_ms - self._queue[0][0]) >= delay:
            self._current_value = self._queue.pop(0)[1]

        return self._current_value


@function('SAMPLE')
class SampleFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._last_value: Optional[float] = None
        self._last_duration: float = 0
        self._last_time_ms: int = 0

    def get_deps(self) -> Set[str]:
        # Function depends on milliseconds only if it's time to reevaluate value
        if time.time() * 1000 - self._last_time_ms >= self._last_duration:
            return {'millisecond'}

        return super().get_deps()

    async def eval(self) -> Evaluated:
        time_ms = int(time.time() * 1000)
        if time_ms - self._last_time_ms < self._last_duration:
            return self._last_value

        self._last_value, self._last_duration = await self.eval_args()
        self._last_time_ms = time_ms

        return self._last_value


@function('FREEZE')
class FreezeFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._last_value: Optional[float] = None
        self._last_duration: float = 0
        self._last_time_ms: int = 0

    def get_deps(self) -> Set[str]:
        # Function depends on milliseconds only when timer is active
        if self._last_time_ms > 0:
            return {'millisecond'}

        return super().get_deps()

    async def eval(self) -> Evaluated:
        time_ms = int(time.time() * 1000)

        if self._last_time_ms == 0:  # Idle
            value = await self.args[0].eval()
            if value != self._last_value:  # Value change detected, start timer
                self._last_time_ms = time_ms
                self._last_duration = await self.args[1].eval()
                self._last_value = value

        else:  # Timer active
            if time_ms - self._last_time_ms > self._last_duration:  # Timer expired
                self._last_time_ms = 0
                return await self.eval()  # Call eval() again, now that _last_time_ms is 0

        return self._last_value


@function('HELD')
class HeldFunction(Function):
    MIN_ARGS = MAX_ARGS = 3

    STATE_OFF = 0
    STATE_WAITING = 1
    STATE_ON = 2

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._time_ms: Optional[int] = None
        self._state = self.STATE_OFF

    def get_deps(self) -> Set[str]:
        if self._state == self.STATE_WAITING:
            return {'millisecond'}

        return super().get_deps()

    async def eval(self) -> Evaluated:
        value, fixed_value, duration = await self.eval_args()

        if value == fixed_value:
            time_ms = int(time.time() * 1000)

            if self._state == self.STATE_OFF:
                self._time_ms = time_ms
                self._state = self.STATE_WAITING

            elif self._state == self.STATE_WAITING:
                delta = time_ms - self._time_ms
                if delta >= duration:
                    self._state = self.STATE_ON

        else:
            self._state = self.STATE_OFF

        return self._state == self.STATE_ON


@function('DERIV')
class DerivFunction(Function):
    MIN_ARGS = MAX_ARGS = 2
    DEPS = ['millisecond']

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._last_value: Optional[float] = None
        self._last_time: float = 0

    async def eval(self) -> Evaluated:
        value, sampling_interval = await self.eval_args()

        sampling_interval /= 1000
        result = 0
        now = time.time()

        if self._last_value is not None:
            delta = now - self._last_time
            if delta < sampling_interval:
                raise EvalSkipped()

            if delta > TIME_JUMP_THRESHOLD:
                self._last_value = value
                self._last_time = now
                raise EvalSkipped()

            else:
                result = (value - self._last_value) / delta

        self._last_value = value
        self._last_time = now

        return result


@function('INTEG')
class IntegFunction(Function):
    MIN_ARGS = MAX_ARGS = 3
    DEPS = ['millisecond']

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._last_value: Optional[float] = None
        self._last_time: float = 0

    async def eval(self) -> Evaluated:
        value, accumulator, sampling_interval = await self.eval_args()

        sampling_interval /= 1000
        result = accumulator
        now = time.time()

        if self._last_value is not None:
            delta = now - self._last_time
            if delta < sampling_interval:
                raise EvalSkipped()

            if delta > TIME_JUMP_THRESHOLD:
                self._last_value = value
                self._last_time = now
                raise EvalSkipped()

            result += (value + self._last_value) * delta / 2

        self._last_value = value
        self._last_time = now

        return result


@function('FMAVG')
class FMAvgFunction(Function):
    MIN_ARGS = MAX_ARGS = 3
    DEPS = ['millisecond']
    QUEUE_SIZE = 1024

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._queue: List[float] = []
        self._last_time: float = 0

    async def eval(self) -> Evaluated:
        value, width, sampling_interval = await self.eval_args()

        width = min(width, self.QUEUE_SIZE)
        sampling_interval /= 1000
        now = time.time()

        if self._last_time > 0:
            delta = now - self._last_time
            if delta < sampling_interval:
                raise EvalSkipped()

            if delta > TIME_JUMP_THRESHOLD:
                self._last_time = now
                raise EvalSkipped()

        # Make place for the new element
        while len(self._queue) >= width:
            self._queue.pop(0)

        self._queue.append(value)
        self._last_time = now

        queue = self._queue[-int(width):]

        return sum(queue) / len(queue)


@function('FMEDIAN')
class FMedianFunction(Function):
    MIN_ARGS = MAX_ARGS = 3
    DEPS = ['millisecond']
    QUEUE_SIZE = 1024

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._queue: List[float] = []
        self._last_time: float = 0

    async def eval(self) -> Evaluated:
        value, width, sampling_interval = await self.eval_args()

        width = min(width, self.QUEUE_SIZE)
        sampling_interval /= 1000
        now = time.time()

        if self._last_time > 0:
            delta = now - self._last_time
            if delta < sampling_interval:
                raise EvalSkipped()

            if delta > TIME_JUMP_THRESHOLD:
                self._last_time = now
                raise EvalSkipped()

        # Make place for the new element
        while len(self._queue) >= width:
            self._queue.pop(0)

        self._queue.append(value)
        self._last_time = now

        queue = self._queue[-int(width):]
        queue.sort()

        return queue[len(queue) // 2]




import asyncio

from typing import Awaitable

from qtoggleserver import startup


def execute(main_code: Awaitable) -> None:
    loop = asyncio.get_event_loop()

    loop.run_until_complete(startup.init_loop())
    loop.run_until_complete(startup.init())

    try:
        run_loop = loop.run_until_complete(main_code)
        if run_loop:
            loop.run_forever()
        loop.run_until_complete(startup.cleanup())

    finally:
        try:
            loop.run_until_complete(startup.cleanup_loop())

        except asyncio.CancelledError:
            pass  # Ignore any cancelled errors

        loop.close()

    startup.logger.info('bye!')




import datetime
import time

from typing import Iterable

import pytz

from qtoggleserver.conf import settings
from qtoggleserver.utils.cmd import run_get_cmd, run_set_cmd


OLD_TIME_LIMIT = 1546304400  # January 2019


class DateError(Exception):
    pass


def has_real_date_time() -> bool:
    # This is to distinguish between unset date (normally around 1970) and a system that has real date time support
    return time.time() > OLD_TIME_LIMIT


def has_set_date_support() -> bool:
    return bool(settings.system.date.set_cmd)


def set_date(date: datetime.datetime) -> None:
    date_str = date.strftime(settings.system.date.set_format)
    run_set_cmd(settings.system.date.set_cmd, cmd_name='date', exc_class=DateError, date=date_str)


def has_timezone_support() -> bool:
    return bool(settings.system.timezone.get_cmd and settings.system.timezone.set_cmd)


def get_timezone() -> str:
    return run_get_cmd(
        settings.system.timezone.get_cmd,
        cmd_name='timezone',
        exc_class=DateError,
        required_fields=['timezone']
    )['timezone']


def set_timezone(timezone: str) -> None:
    run_set_cmd(settings.system.timezone.set_cmd, cmd_name='timezone', exc_class=DateError, timezone=timezone)


def get_timezones() -> Iterable[str]:
    return pytz.common_timezones




import math

from .base import Evaluated
from .functions import function, Function


@function('FLOOR')
class FloorFunction(Function):
    MIN_ARGS = MAX_ARGS = 1

    async def eval(self) -> Evaluated:
        eval_args = await self.eval_args()

        return int(math.floor(eval_args[0]))


@function('CEIL')
class CeilFunction(Function):
    MIN_ARGS = MAX_ARGS = 1

    async def eval(self) -> Evaluated:
        eval_args = await self.eval_args()

        return int(math.ceil(eval_args[0]))


@function('ROUND')
class RoundFunction(Function):
    MIN_ARGS = 1
    MAX_ARGS = 2

    async def eval(self) -> Evaluated:
        eval_args = await self.eval_args()

        v = eval_args[0]
        d = 0
        if len(eval_args) == 2:
            d = eval_args[1]

        return round(v, int(d))




import copy
import logging
import operator
import os

from typing import Any, Dict, Iterable, List, Optional, Tuple

from qtoggleserver.conf import settings
from qtoggleserver.persist import BaseDriver
from qtoggleserver.persist.typing import Id, Record
from qtoggleserver.utils import json as json_utils


DEFAULT_FILE_PATH = 'qtoggleserver-data.json'

FILTER_OP_MAPPING = {
    'gt': operator.gt,
    'ge': operator.ge,
    'lt': operator.lt,
    'le': operator.le,
    'in': lambda a, b: a in b
}


logger = logging.getLogger(__name__)

Collection = Dict[int, Record]
IndexedData = Dict[str, Collection]
UnindexedData = Dict[str, List[Record]]


class JSONPersistError(Exception):
    pass


class DuplicateRecordId(JSONPersistError):
    pass


class JSONDriver(BaseDriver):
    def __init__(
        self,
        file_path: str = DEFAULT_FILE_PATH,
        pretty_format: Optional[bool] = None,
        use_backup: bool = True,
        **kwargs
    ) -> None:
        logger.debug('using file %s', file_path)

        # Pretty formatting follows general debug flag, by default
        if pretty_format is None:
            pretty_format = settings.debug

        self._file_path: str = file_path
        self._pretty_format: bool = pretty_format
        self._use_backup: bool = use_backup

        self._data: IndexedData = self._index(self._load())

    async def query(
        self,
        collection: str,
        fields: Optional[List[str]],
        filt: Dict[str, Any],
        sort: List[Tuple[str, bool]],
        limit: Optional[int]
    ) -> Iterable[Record]:

        coll = self._data.get(collection, {})
        records = []

        if isinstance(filt.get('id'), Id):  # Look for specific record id
            filt = dict(filt)
            id_ = filt.pop('id')
            record = coll.get(id_)

            # Apply filter criteria
            if record is not None and self._filter_matches(record, filt):
                records.append(record)

        else:  # No single specific id in filt
            for id_, record in coll.items():
                # Apply filter criteria
                if self._filter_matches(dict(record, id=id_), filt):
                    records.append(record)

        # Sort
        for field, rev in reversed(sort):
            if field == 'id':
                records.sort(key=lambda r: int(r['id']), reverse=rev)

            else:
                records.sort(key=lambda r: r.get(field), reverse=rev)

        # Apply limit
        if limit is not None:
            records = records[:limit]

        # Apply projection
        if fields is not None:
            fields = set(fields)
            projected_records = []
            for record in records:
                projected_record = {k: v for k, v in record.items() if k in fields}
                projected_records.append(projected_record)

            records = projected_records

        return copy.deepcopy(records)

    async def insert(self, collection: str, record: Record) -> Id:
        coll = self._data.setdefault(collection, {})

        id_ = record.get('id')
        if id_ is None:
            id_ = self._find_next_id(coll)
            record = dict(record, id=id_)

        elif id_ in coll:
            raise DuplicateRecordId(id_)

        coll[id_] = record

        self._save(self._unindex(self._data))

        return id_

    async def update(self, collection: str, record_part: Record, filt: Dict[str, Any]) -> int:
        coll = self._data.setdefault(collection, {})
        modified_count = 0

        if isinstance(filt.get('id'), Id):
            filt = dict(filt)
            id_ = filt.pop('id')

            record = coll.get(id_)
            if record is not None:
                record.update(record_part)
                modified_count = 1

        else:  # No single specific id in filt
            for id_, record in coll.items():
                # Apply filter criteria
                if not self._filter_matches(dict(record, id=id_), filt):
                    continue

                # Actually update the record
                record.update(record_part)
                modified_count += 1

        self._save(self._unindex(self._data))

        return modified_count

    async def replace(self, collection: str, id_: Id, record: Record) -> bool:
        coll = self._data.setdefault(collection, {})

        if coll.get(id_) is None:
            return False  # No record found, no replacing

        record = dict(record)

        # Never change record id with replace
        record['id'] = id_
        coll[id_] = record

        self._save(self._unindex(self._data))

        return True

    async def remove(self, collection: str, filt: Dict[str, Any]) -> int:
        coll = self._data.setdefault(collection, {})
        removed_count = 0

        if isinstance(filt.get('id'), Id):
            filt = dict(filt)
            id_ = filt.pop('id')

            record = coll.get(id_)
            if (record is not None) and self._filter_matches(record, filt):
                coll.pop(id_)
                removed_count = 1

        else:  # No single specific id in filt
            for id_, record in list(coll.items()):
                # Apply filter criteria
                if not self._filter_matches(dict(record, id=id_), filt):
                    continue

                # Actually remove the record
                coll.pop(id_)
                removed_count += 1

        self._save(self._unindex(self._data))

        return removed_count

    def _filter_matches(self, record: Record, filt: Dict[str, Any]) -> bool:
        for key, value in filt.items():
            try:
                db_record_value = record[key]

            except KeyError:
                return False

            if not self._filter_value_matches(db_record_value, value):
                return False

        return True

    @staticmethod
    def _filter_value_matches(db_record_value: Any, filt_value: Any) -> bool:
        if isinstance(filt_value, dict):  # filter with operators
            for op, v in filt_value.items():
                op_func = FILTER_OP_MAPPING[op]
                if not op_func(db_record_value, v):
                    return False

            return True

        else:  # Assuming simple value
            return db_record_value == filt_value

    @staticmethod
    def _find_next_id(coll: Collection) -> Id:
        int_ids = [0]
        for id_ in coll.keys():
            try:
                int_ids.append(int(id_))

            except ValueError:
                continue

        return str(max(int_ids) + 1)

    def _get_backup_file_path(self) -> str:
        path, ext = os.path.splitext(self._file_path)
        return f'{path}_backup{ext}'

    def _load(self) -> UnindexedData:
        logger.debug('loading from %s', self._file_path)

        try:
            # If the file is accessible but empty, consider data loaded and return empty dictionary
            if os.stat(self._file_path).st_size == 0:
                logger.debug('file %s is empty', self._file_path)
                return {}

        except FileNotFoundError:
            # If the file does not exist, consider data loaded and return empty dictionary
            logger.debug('file %s does not exist', self._file_path)
            return {}

        try:
            with open(self._file_path, 'rb') as f:
                data = f.read()
                return json_utils.loads(data, allow_extended_types=True)

        except Exception as e:
            if not self._use_backup:
                raise

            # Upon failure, if using a backup, simply log the error and attempt to load from backup file
            logger.error('failed to load from %s: %s', self._file_path, e, exc_info=True)

            backup_file_path = self._get_backup_file_path()
            logger.warning('loading from backup %s', backup_file_path)

            with open(backup_file_path, 'rb') as f:
                return json_utils.loads(f.read(), allow_extended_types=True)

    def _save(self, data: UnindexedData) -> None:
        if self._use_backup and os.path.exists(self._file_path):
            backup_file_path = self._get_backup_file_path()
            logger.debug('backing up %s to %s', self._file_path, backup_file_path)
            os.rename(self._file_path, backup_file_path)

        logger.debug('saving to %s', self._file_path)

        with open(self._file_path, 'wb') as f:
            data = json_utils.dumps(data, allow_extended_types=True, indent=4 if self._pretty_format else None)
            f.write(data.encode())

    @staticmethod
    def _index(data: UnindexedData) -> IndexedData:
        indexed_data = {}
        for coll, records in data.items():
            indexed_data[coll] = {r.get('id', ''): r for r in records}

        return indexed_data

    @staticmethod
    def _unindex(data: IndexedData) -> UnindexedData:
        unindexed_data = {}
        for coll, records in data.items():
            unindexed_data[coll] = list(records.values())

        return unindexed_data




import asyncio
import logging

from typing import Callable, List, Optional

from qtoggleserver.core.typing import PortValue


logger = logging.getLogger(__name__)


class SequenceError(Exception):
    pass


class Sequence:
    def __init__(
        self,
        values: List[PortValue],
        delays: List[int],
        repeat: int,
        callback: Callable,
        finish_callback: Callable
    ) -> None:

        self._values: List[PortValue] = values
        self._delays: List[int] = delays
        self._repeat: int = repeat

        self._callback: Callable = callback
        self._finish_callback: Callable = finish_callback
        self._counter: int = 0
        self._loop_task: Optional[asyncio.Task] = None

    def start(self) -> None:
        if self._loop_task:
            raise SequenceError('Loop task already started')

        self._loop_task = asyncio.create_task(self._loop())

    async def cancel(self) -> None:
        if self._loop_task:
            self._loop_task.cancel()
            await self._loop_task

    async def _loop(self) -> None:
        for i, value in enumerate(self._values):
            try:
                try:
                    self._callback(value)

                except Exception as e:
                    logger.error('sequence callback failed: %s', e, exc_info=True)

                if i < len(self._values) - 1:
                    await asyncio.sleep(self._delays[i] / 1000.0)

                else:
                    if self._repeat > 0 and self._counter >= self._repeat - 1:
                        await self._finish_callback()

                        return

                    self._counter += 1
                    await asyncio.sleep(self._delays[i] / 1000.0)

                    self._loop_task = asyncio.create_task(self._loop())
                    return

            except asyncio.CancelledError:
                logger.debug('sequence loop cancelled')
                break

        self._loop_task = None
